"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import Optional, TYPE_CHECKING


from .asset import Asset
from .enums import NameplatePalette, CollectibleType, try_enum
from .utils import parse_time


if TYPE_CHECKING:
    from datetime import datetime

    from .state import ConnectionState
    from .types.user import (
        Collectible as CollectiblePayload,
    )


__all__ = ('Collectible',)


class Collectible:
    """Represents a user's collectible.

    .. versionadded:: 2.7

    Attributes
    ----------
    label: :class:`str`
        The label of the collectible.
    palette: Optional[:class:`NameplatePalette`]
        The palette of the collectible.
        This is only available if ``type`` is
        :class:`CollectibleType.nameplate`.
    sku_id: :class:`int`
        The SKU ID of the collectible.
    type: :class:`CollectibleType`
        The type of the collectible.
    expires_at: Optional[:class:`datetime.datetime`]
        The expiration date of the collectible. If applicable.
    """

    __slots__ = (
        'type',
        'sku_id',
        'label',
        'expires_at',
        'palette',
        '_state',
        '_asset',
    )

    def __init__(self, *, state: ConnectionState, type: str, data: CollectiblePayload) -> None:
        self._state: ConnectionState = state
        self.type: CollectibleType = try_enum(CollectibleType, type)
        self._asset: str = data['asset']
        self.sku_id: int = int(data['sku_id'])
        self.label: str = data['label']
        self.expires_at: Optional[datetime] = parse_time(data.get('expires_at'))

        # nameplate
        self.palette: Optional[NameplatePalette]
        try:
            self.palette = try_enum(NameplatePalette, data['palette'])  # type: ignore
        except KeyError:
            self.palette = None

    @property
    def static(self) -> Asset:
        """:class:`Asset`: The static asset of the collectible."""
        return Asset._from_user_collectible(self._state, self._asset)

    @property
    def animated(self) -> Asset:
        """:class:`Asset`: The animated asset of the collectible."""
        return Asset._from_user_collectible(self._state, self._asset, animated=True)

    def __repr__(self) -> str:
        attrs = ['sku_id']
        if self.palette:
            attrs.append('palette')

        joined_attrs = ' '.join(f'{attr}={getattr(self, attr)!r}' for attr in attrs)
        return f'<{self.type.name.title()} {joined_attrs}>'



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Literal, TypedDict, Union, Optional
from typing_extensions import NotRequired

from .channel import (
    ChannelTypeWithoutThread,
    GuildChannel,
    InteractionDMChannel,
    GroupDMChannel,
)
from .sku import Entitlement
from .threads import ThreadType, ThreadMetadata
from .member import Member
from .message import Attachment
from .role import Role
from .snowflake import Snowflake
from .user import User
from .guild import GuildFeature
from .components import ComponentBase

if TYPE_CHECKING:
    from .message import Message


InteractionType = Literal[1, 2, 3, 4, 5]
InteractionResponseType = Literal[
    1,
    4,
    5,
    6,
    7,
    8,
    9,
    10,
]
InteractionContextType = Literal[0, 1, 2]
InteractionInstallationType = Literal[0, 1]


class _BasePartialChannel(TypedDict):
    id: Snowflake
    name: str
    permissions: str


class PartialChannel(_BasePartialChannel):
    type: ChannelTypeWithoutThread
    topic: NotRequired[str]
    position: int
    nsfw: bool
    flags: int
    rate_limit_per_user: int
    parent_id: Optional[Snowflake]
    last_message_id: Optional[Snowflake]
    last_pin_timestamp: NotRequired[str]


class PartialThread(_BasePartialChannel):
    type: ThreadType
    thread_metadata: ThreadMetadata
    parent_id: Snowflake
    applied_tags: NotRequired[List[Snowflake]]
    owner_id: Snowflake
    message_count: int
    member_count: int
    rate_limit_per_user: int
    last_message_id: NotRequired[Optional[Snowflake]]
    flags: NotRequired[int]
    total_message_sent: int


class ResolvedData(TypedDict, total=False):
    users: Dict[str, User]
    members: Dict[str, Member]
    roles: Dict[str, Role]
    channels: Dict[str, Union[PartialChannel, PartialThread]]
    messages: Dict[str, Message]
    attachments: Dict[str, Attachment]


class PartialInteractionGuild(TypedDict):
    id: Snowflake
    locale: str
    features: List[GuildFeature]


class _BaseApplicationCommandInteractionDataOption(TypedDict):
    name: str


class _CommandGroupApplicationCommandInteractionDataOption(_BaseApplicationCommandInteractionDataOption):
    type: Literal[1, 2]
    options: List[ApplicationCommandInteractionDataOption]


class _BaseValueApplicationCommandInteractionDataOption(_BaseApplicationCommandInteractionDataOption, total=False):
    focused: bool


class _StringValueApplicationCommandInteractionDataOption(_BaseValueApplicationCommandInteractionDataOption):
    type: Literal[3]
    value: str


class _IntegerValueApplicationCommandInteractionDataOption(_BaseValueApplicationCommandInteractionDataOption):
    type: Literal[4]
    value: int


class _BooleanValueApplicationCommandInteractionDataOption(_BaseValueApplicationCommandInteractionDataOption):
    type: Literal[5]
    value: bool


class _SnowflakeValueApplicationCommandInteractionDataOption(_BaseValueApplicationCommandInteractionDataOption):
    type: Literal[6, 7, 8, 9, 11]
    value: Snowflake


class _NumberValueApplicationCommandInteractionDataOption(_BaseValueApplicationCommandInteractionDataOption):
    type: Literal[10]
    value: float


_ValueApplicationCommandInteractionDataOption = Union[
    _StringValueApplicationCommandInteractionDataOption,
    _IntegerValueApplicationCommandInteractionDataOption,
    _BooleanValueApplicationCommandInteractionDataOption,
    _SnowflakeValueApplicationCommandInteractionDataOption,
    _NumberValueApplicationCommandInteractionDataOption,
]


ApplicationCommandInteractionDataOption = Union[
    _CommandGroupApplicationCommandInteractionDataOption,
    _ValueApplicationCommandInteractionDataOption,
]


class _BaseApplicationCommandInteractionData(TypedDict):
    id: Snowflake
    name: str
    resolved: NotRequired[ResolvedData]
    guild_id: NotRequired[Snowflake]


class ChatInputApplicationCommandInteractionData(_BaseApplicationCommandInteractionData, total=False):
    type: Literal[1]
    options: List[ApplicationCommandInteractionDataOption]


class _BaseNonChatInputApplicationCommandInteractionData(_BaseApplicationCommandInteractionData):
    target_id: Snowflake


class UserApplicationCommandInteractionData(_BaseNonChatInputApplicationCommandInteractionData):
    type: Literal[2]


class MessageApplicationCommandInteractionData(_BaseNonChatInputApplicationCommandInteractionData):
    type: Literal[3]


ApplicationCommandInteractionData = Union[
    ChatInputApplicationCommandInteractionData,
    UserApplicationCommandInteractionData,
    MessageApplicationCommandInteractionData,
]


class _BaseMessageComponentInteractionData(TypedDict):
    custom_id: str


class ButtonMessageComponentInteractionData(_BaseMessageComponentInteractionData):
    component_type: Literal[2]


class SelectMessageComponentInteractionData(_BaseMessageComponentInteractionData):
    component_type: Literal[3, 5, 6, 7, 8]
    values: List[str]
    resolved: NotRequired[ResolvedData]


MessageComponentInteractionData = Union[ButtonMessageComponentInteractionData, SelectMessageComponentInteractionData]


class ModalSubmitTextInputInteractionData(ComponentBase):
    type: Literal[4]
    custom_id: str
    value: str


class ModalSubmitSelectInteractionData(ComponentBase):
    type: Literal[3, 5, 6, 7, 8]
    custom_id: str
    values: List[str]


class ModalSubmitFileUploadInteractionData(ComponentBase):
    type: Literal[19]
    custom_id: str
    values: List[str]


class ModalSubmitRadioGroupInteractionData(ComponentBase):
    type: Literal[21]
    custom_id: str
    id: int
    value: Optional[str]


class ModalSubmitCheckboxGroupInteractionData(ComponentBase):
    type: Literal[22]
    custom_id: str
    id: int
    values: List[str]


class ModalSubmitCheckboxInteractionData(ComponentBase):
    type: Literal[23]
    custom_id: str
    id: int
    value: bool


ModalSubmitLabelComponentItemInteractionData = Union[
    ModalSubmitSelectInteractionData,
    ModalSubmitTextInputInteractionData,
    ModalSubmitFileUploadInteractionData,
    ModalSubmitRadioGroupInteractionData,
    ModalSubmitCheckboxGroupInteractionData,
    ModalSubmitCheckboxInteractionData,
]


class ModalSubmitActionRowInteractionData(TypedDict):
    type: Literal[1]
    components: List[ModalSubmitTextInputInteractionData]


class ModalSubmitTextDisplayInteractionData(ComponentBase):
    type: Literal[10]
    content: str


class ModalSubmitLabelInteractionData(ComponentBase):
    type: Literal[18]
    component: ModalSubmitLabelComponentItemInteractionData


ModalSubmitComponentInteractionData = Union[
    ModalSubmitActionRowInteractionData,
    ModalSubmitTextDisplayInteractionData,
    ModalSubmitLabelInteractionData,
]


class ModalSubmitInteractionData(TypedDict):
    custom_id: str
    components: List[ModalSubmitComponentInteractionData]
    resolved: NotRequired[ResolvedData]


InteractionData = Union[
    ApplicationCommandInteractionData,
    MessageComponentInteractionData,
    ModalSubmitInteractionData,
]


class _BaseInteraction(TypedDict):
    id: Snowflake
    application_id: Snowflake
    token: str
    version: Literal[1]
    guild_id: NotRequired[Snowflake]
    guild: NotRequired[PartialInteractionGuild]
    channel_id: NotRequired[Snowflake]
    channel: Union[GuildChannel, InteractionDMChannel, GroupDMChannel]
    app_permissions: NotRequired[str]
    locale: NotRequired[str]
    guild_locale: NotRequired[str]
    entitlement_sku_ids: NotRequired[List[Snowflake]]
    entitlements: NotRequired[List[Entitlement]]
    authorizing_integration_owners: Dict[Literal['0', '1'], Snowflake]
    context: NotRequired[InteractionContextType]
    attachment_size_limit: int


class PingInteraction(_BaseInteraction):
    type: Literal[1]


class ApplicationCommandInteraction(_BaseInteraction):
    type: Literal[2, 4]
    data: ApplicationCommandInteractionData


class MessageComponentInteraction(_BaseInteraction):
    type: Literal[3]
    data: MessageComponentInteractionData


class ModalSubmitInteraction(_BaseInteraction):
    type: Literal[5]
    data: ModalSubmitInteractionData


Interaction = Union[
    PingInteraction,
    ApplicationCommandInteraction,
    MessageComponentInteraction,
    ModalSubmitInteraction,
]


class MessageInteraction(TypedDict):
    id: Snowflake
    type: InteractionType
    name: str
    user: User
    member: NotRequired[Member]


class _MessageInteractionMetadata(TypedDict):
    id: Snowflake
    user: User
    authorizing_integration_owners: Dict[Literal['0', '1'], Snowflake]
    original_response_message_id: NotRequired[Snowflake]


class _ApplicationCommandMessageInteractionMetadata(_MessageInteractionMetadata):
    type: Literal[2]
    # command_type: Literal[1, 2, 3, 4]


class UserApplicationCommandMessageInteractionMetadata(_ApplicationCommandMessageInteractionMetadata):
    # command_type: Literal[2]
    target_user: User


class MessageApplicationCommandMessageInteractionMetadata(_ApplicationCommandMessageInteractionMetadata):
    # command_type: Literal[3]
    target_message_id: Snowflake


ApplicationCommandMessageInteractionMetadata = Union[
    _ApplicationCommandMessageInteractionMetadata,
    UserApplicationCommandMessageInteractionMetadata,
    MessageApplicationCommandMessageInteractionMetadata,
]


class MessageComponentMessageInteractionMetadata(_MessageInteractionMetadata):
    type: Literal[3]
    interacted_message_id: Snowflake


class ModalSubmitMessageInteractionMetadata(_MessageInteractionMetadata):
    type: Literal[5]
    triggering_interaction_metadata: Union[
        ApplicationCommandMessageInteractionMetadata,
        MessageComponentMessageInteractionMetadata,
    ]


MessageInteractionMetadata = Union[
    ApplicationCommandMessageInteractionMetadata,
    MessageComponentMessageInteractionMetadata,
    ModalSubmitMessageInteractionMetadata,
]


class InteractionCallbackResponse(TypedDict):
    id: Snowflake
    type: InteractionType
    activity_instance_id: NotRequired[str]
    response_message_id: NotRequired[Snowflake]
    response_message_loading: NotRequired[bool]
    response_message_ephemeral: NotRequired[bool]


class InteractionCallbackActivity(TypedDict):
    id: str


class InteractionCallbackResource(TypedDict):
    type: InteractionResponseType
    activity_instance: NotRequired[InteractionCallbackActivity]
    message: NotRequired[Message]


class InteractionCallback(TypedDict):
    interaction: InteractionCallbackResponse
    resource: NotRequired[InteractionCallbackResource]



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from .snowflake import Snowflake
from typing import Literal, Optional, TypedDict
from typing_extensions import NotRequired


PremiumType = Literal[0, 1, 2, 3]
NameplatePallete = Literal['crimson', 'berry', 'sky', 'teal', 'forest', 'bubble_gum', 'violet', 'cobalt', 'clover']


class _UserSKU(TypedDict):
    asset: str
    sku_id: Snowflake


AvatarDecorationData = _UserSKU


class PrimaryGuild(TypedDict):
    identity_guild_id: Optional[int]
    identity_enabled: Optional[bool]
    tag: Optional[str]
    badge: Optional[str]


class Collectible(_UserSKU):
    label: str
    expires_at: Optional[str]


class NameplateCollectible(Collectible):
    palette: str


class UserCollectibles(TypedDict):
    nameplate: NameplateCollectible


class PartialUser(TypedDict):
    id: Snowflake
    username: str
    discriminator: str
    avatar: Optional[str]
    global_name: Optional[str]
    avatar_decoration_data: NotRequired[AvatarDecorationData]
    primary_guild: NotRequired[PrimaryGuild]
    collectibles: NotRequired[UserCollectibles]


class User(PartialUser, total=False):
    bot: bool
    system: bool
    mfa_enabled: bool
    locale: str
    verified: bool
    email: Optional[str]
    flags: int
    premium_type: PremiumType
    public_flags: int



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import colorsys
import random
import re

from typing import TYPE_CHECKING, Optional, Tuple, Union

if TYPE_CHECKING:
    from typing_extensions import Self

__all__ = (
    'Colour',
    'Color',
)

RGB_REGEX = re.compile(r'rgb\s*\((?P<r>[0-9.]+%?)\s*,\s*(?P<g>[0-9.]+%?)\s*,\s*(?P<b>[0-9.]+%?)\s*\)')


def parse_hex_number(argument: str) -> Colour:
    arg = ''.join(i * 2 for i in argument) if len(argument) == 3 else argument
    try:
        value = int(arg, base=16)
        if not (0 <= value <= 0xFFFFFF):
            raise ValueError('hex number out of range for 24-bit colour')
    except ValueError:
        raise ValueError('invalid hex digit given') from None
    else:
        return Color(value=value)


def parse_rgb_number(number: str) -> int:
    if number[-1] == '%':
        value = float(number[:-1])
        if not (0 <= value <= 100):
            raise ValueError('rgb percentage can only be between 0 to 100')
        return round(255 * (value / 100))

    value = int(number)
    if not (0 <= value <= 255):
        raise ValueError('rgb number can only be between 0 to 255')
    return value


def parse_rgb(argument: str, *, regex: re.Pattern[str] = RGB_REGEX) -> Colour:
    match = regex.match(argument)
    if match is None:
        raise ValueError('invalid rgb syntax found')

    red = parse_rgb_number(match.group('r'))
    green = parse_rgb_number(match.group('g'))
    blue = parse_rgb_number(match.group('b'))
    return Color.from_rgb(red, green, blue)


class Colour:
    """Represents a Discord role colour. This class is similar
    to a (red, green, blue) :class:`tuple`.

    There is an alias for this called Color.

    .. container:: operations

        .. describe:: x == y

             Checks if two colours are equal.

        .. describe:: x != y

             Checks if two colours are not equal.

        .. describe:: hash(x)

             Return the colour's hash.

        .. describe:: str(x)

             Returns the hex format for the colour.

        .. describe:: int(x)

             Returns the raw colour value.

    .. note::

        The colour values in the classmethods are mostly provided as-is and can change between
        versions should the Discord client's representation of that colour also change.

    Attributes
    ------------
    value: :class:`int`
        The raw integer colour value.
    """

    __slots__ = ('value',)

    def __init__(self, value: int):
        if not isinstance(value, int):
            raise TypeError(f'Expected int parameter, received {value.__class__.__name__} instead.')

        self.value: int = value

    def _get_byte(self, byte: int) -> int:
        return (self.value >> (8 * byte)) & 0xFF

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Colour) and self.value == other.value

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return f'#{self.value:0>6x}'

    def __int__(self) -> int:
        return self.value

    def __repr__(self) -> str:
        return f'<Colour value={self.value}>'

    def __hash__(self) -> int:
        return hash(self.value)

    @property
    def r(self) -> int:
        """:class:`int`: Returns the red component of the colour."""
        return self._get_byte(2)

    @property
    def g(self) -> int:
        """:class:`int`: Returns the green component of the colour."""
        return self._get_byte(1)

    @property
    def b(self) -> int:
        """:class:`int`: Returns the blue component of the colour."""
        return self._get_byte(0)

    def to_rgb(self) -> Tuple[int, int, int]:
        """Tuple[:class:`int`, :class:`int`, :class:`int`]: Returns an (r, g, b) tuple representing the colour."""
        return (self.r, self.g, self.b)

    @classmethod
    def from_rgb(cls, r: int, g: int, b: int) -> Self:
        """Constructs a :class:`Colour` from an RGB tuple."""
        return cls((r << 16) + (g << 8) + b)

    @classmethod
    def from_hsv(cls, h: float, s: float, v: float) -> Self:
        """Constructs a :class:`Colour` from an HSV tuple."""
        rgb = colorsys.hsv_to_rgb(h, s, v)
        return cls.from_rgb(*(int(x * 255) for x in rgb))

    @classmethod
    def from_str(cls, value: str) -> Colour:
        """Constructs a :class:`Colour` from a string.

        The following formats are accepted:

        - ``0x<hex>``
        - ``#<hex>``
        - ``0x#<hex>``
        - ``rgb(<number>, <number>, <number>)``

        Like CSS, ``<number>`` can be either 0-255 or 0-100% and ``<hex>`` can be
        either a 6 digit hex number or a 3 digit hex shortcut (e.g. #FFF).

        .. versionadded:: 2.0

        Raises
        -------
        ValueError
            The string could not be converted into a colour.
        """

        if not value:
            raise ValueError('unknown colour format given')

        if value[0] == '#':
            return parse_hex_number(value[1:])

        if value[0:2] == '0x':
            rest = value[2:]
            # Legacy backwards compatible syntax
            if rest.startswith('#'):
                return parse_hex_number(rest[1:])
            return parse_hex_number(rest)

        arg = value.lower()
        if arg[0:3] == 'rgb':
            return parse_rgb(arg)

        raise ValueError('unknown colour format given')

    @classmethod
    def default(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0``.

        .. colour:: #000000
        """
        return cls(0)

    @classmethod
    def random(cls, *, seed: Optional[Union[int, str, float, bytes, bytearray]] = None) -> Self:
        """A factory method that returns a :class:`Colour` with a random hue.

        .. note::

            The random algorithm works by choosing a colour with a random hue but
            with maxed out saturation and value.

        .. versionadded:: 1.6

        Parameters
        ------------
        seed: Optional[Union[:class:`int`, :class:`str`, :class:`float`, :class:`bytes`, :class:`bytearray`]]
            The seed to initialize the RNG with. If ``None`` is passed the default RNG is used.

            .. versionadded:: 1.7
        """
        rand = random if seed is None else random.Random(seed)
        return cls.from_hsv(rand.random(), 1, 1)

    @classmethod
    def teal(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x1ABC9C``.

        .. colour:: #1ABC9C
        """
        return cls(0x1ABC9C)

    @classmethod
    def dark_teal(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x11806A``.

        .. colour:: #11806A
        """
        return cls(0x11806A)

    @classmethod
    def brand_green(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x57F287``.

        .. colour:: #57F287


        .. versionadded:: 2.0
        """
        return cls(0x57F287)

    @classmethod
    def green(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x2ECC71``.

        .. colour:: #2ECC71
        """
        return cls(0x2ECC71)

    @classmethod
    def dark_green(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x1F8B4C``.

        .. colour:: #1F8B4C
        """
        return cls(0x1F8B4C)

    @classmethod
    def blue(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x3498DB``.

        .. colour:: #3498DB
        """
        return cls(0x3498DB)

    @classmethod
    def dark_blue(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x206694``.

        .. colour:: #206694
        """
        return cls(0x206694)

    @classmethod
    def purple(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x9B59B6``.

        .. colour:: #9B59B6
        """
        return cls(0x9B59B6)

    @classmethod
    def dark_purple(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x71368A``.

        .. colour:: #71368A
        """
        return cls(0x71368A)

    @classmethod
    def magenta(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xE91E63``.

        .. colour:: #E91E63
        """
        return cls(0xE91E63)

    @classmethod
    def dark_magenta(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xAD1457``.

        .. colour:: #AD1457
        """
        return cls(0xAD1457)

    @classmethod
    def gold(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xF1C40F``.

        .. colour:: #F1C40F
        """
        return cls(0xF1C40F)

    @classmethod
    def dark_gold(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xC27C0E``.

        .. colour:: #C27C0E
        """
        return cls(0xC27C0E)

    @classmethod
    def orange(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xE67E22``.

        .. colour:: #E67E22
        """
        return cls(0xE67E22)

    @classmethod
    def dark_orange(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xA84300``.

        .. colour:: #A84300
        """
        return cls(0xA84300)

    @classmethod
    def brand_red(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xED4245``.

        .. colour:: #ED4245

        .. versionadded:: 2.0
        """
        return cls(0xED4245)

    @classmethod
    def red(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xE74C3C``.

        .. colour:: #E74C3C
        """
        return cls(0xE74C3C)

    @classmethod
    def dark_red(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x992D22``.

        .. colour:: #992D22
        """
        return cls(0x992D22)

    @classmethod
    def lighter_grey(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x95A5A6``.

        .. colour:: #95A5A6
        """
        return cls(0x95A5A6)

    lighter_gray = lighter_grey

    @classmethod
    def dark_grey(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x607d8b``.

        .. colour:: #607d8b
        """
        return cls(0x607D8B)

    dark_gray = dark_grey

    @classmethod
    def light_grey(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x979C9F``.

        .. colour:: #979C9F
        """
        return cls(0x979C9F)

    light_gray = light_grey

    @classmethod
    def darker_grey(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x546E7A``.

        .. colour:: #546E7A
        """
        return cls(0x546E7A)

    darker_gray = darker_grey

    @classmethod
    def og_blurple(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x7289DA``.

        .. colour:: #7289DA
        """
        return cls(0x7289DA)

    @classmethod
    def blurple(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x5865F2``.

        .. colour:: #5865F2
        """
        return cls(0x5865F2)

    @classmethod
    def greyple(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x99AAB5``.

        .. colour:: #99AAB5
        """
        return cls(0x99AAB5)

    @classmethod
    def ash_theme(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x2E2E34``.

        This will appear transparent on Discord's ash theme.

        .. colour:: #2E2E34

        .. versionadded:: 2.6
        """
        return cls(0x2E2E34)

    @classmethod
    def dark_theme(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x1A1A1E``.

        This will appear transparent on Discord's dark theme.

        .. colour:: #1A1A1E

        .. versionadded:: 1.5

        .. versionchanged:: 2.2
            Updated colour from previous ``0x36393F`` to reflect discord theme changes.

        .. versionchanged:: 2.6
            Updated colour from previous ``0x313338`` to reflect discord theme changes.
        """
        return cls(0x1A1A1E)

    @classmethod
    def onyx_theme(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x070709``.

        This will appear transparent on Discord's onyx theme.

        .. colour:: #070709

        .. versionadded:: 2.6
        """
        return cls(0x070709)

    @classmethod
    def light_theme(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xFBFBFB``.

        This will appear transparent on Discord's light theme.

        .. colour:: #FBFBFB

        .. versionadded:: 2.6
        """
        return cls(0xFBFBFB)

    @classmethod
    def fuchsia(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xEB459E``.

        .. colour:: #EB459E

        .. versionadded:: 2.0
        """
        return cls(0xEB459E)

    @classmethod
    def yellow(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xFEE75C``.

        .. colour:: #FEE75C

        .. versionadded:: 2.0
        """
        return cls(0xFEE75C)

    @classmethod
    def ash_embed(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x37373E``.

        .. colour:: #37373E

        .. versionadded:: 2.6

        """
        return cls(0x37373E)

    @classmethod
    def dark_embed(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x242429``.

        .. colour:: #242429

        .. versionadded:: 2.2

        .. versionchanged:: 2.6
            Updated colour from previous ``0x2B2D31`` to reflect discord theme changes.
        """
        return cls(0x242429)

    @classmethod
    def onyx_embed(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0x131416``.

        .. colour:: #131416

        .. versionadded:: 2.6
        """
        return cls(0x131416)

    @classmethod
    def light_embed(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xFFFFFF``.

        .. colour:: #EEEFF1

        .. versionadded:: 2.2

        .. versionchanged:: 2.6
            Updated colour from previous ``0xEEEFF1`` to reflect discord theme changes.
        """
        return cls(0xFFFFFF)

    @classmethod
    def pink(cls) -> Self:
        """A factory method that returns a :class:`Colour` with a value of ``0xEB459F``.

        .. colour:: #EB459F

        .. versionadded:: 2.3
        """
        return cls(0xEB459F)


Color = Colour



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import Literal, Optional, TypedDict, Union
from typing_extensions import NotRequired

from .scheduled_event import GuildScheduledEvent
from .snowflake import Snowflake
from .guild import InviteGuild, _GuildPreviewUnique
from .channel import PartialChannel
from .user import PartialUser
from .appinfo import PartialAppInfo

InviteTargetType = Literal[1, 2]
InviteType = Literal[0, 1, 2]


class _InviteMetadata(TypedDict, total=False):
    uses: int
    max_uses: int
    max_age: int
    temporary: bool
    created_at: str


class VanityInvite(_InviteMetadata):
    code: Optional[str]
    revoked: NotRequired[bool]


class IncompleteInvite(_InviteMetadata):
    code: str
    channel: PartialChannel


class Invite(IncompleteInvite, total=False):
    guild: InviteGuild
    inviter: PartialUser
    target_user: PartialUser
    target_type: InviteTargetType
    target_application: PartialAppInfo
    guild_scheduled_event: GuildScheduledEvent
    type: InviteType
    flags: NotRequired[int]
    expires_at: Optional[str]


class InviteWithCounts(Invite, _GuildPreviewUnique): ...


class GatewayInviteCreate(TypedDict):
    channel_id: Snowflake
    code: str
    created_at: str
    expires_at: Optional[str]
    max_age: int
    max_uses: int
    temporary: bool
    uses: bool
    guild_id: Snowflake
    inviter: NotRequired[PartialUser]
    target_type: NotRequired[InviteTargetType]
    target_user: NotRequired[PartialUser]
    target_application: NotRequired[PartialAppInfo]
    flags: NotRequired[int]


class GatewayInviteDelete(TypedDict):
    channel_id: Snowflake
    code: str
    guild_id: NotRequired[Snowflake]


GatewayInvite = Union[GatewayInviteCreate, GatewayInviteDelete]



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import types
from collections import namedtuple
from typing import Any, ClassVar, Dict, List, Optional, TYPE_CHECKING, Tuple, Type, TypeVar, Iterator, Mapping

__all__ = (
    'Enum',
    'ChannelType',
    'MessageType',
    'SpeakingState',
    'VerificationLevel',
    'ContentFilter',
    'Status',
    'DefaultAvatar',
    'AuditLogAction',
    'AuditLogActionCategory',
    'UserFlags',
    'ActivityType',
    'NotificationLevel',
    'TeamMembershipState',
    'TeamMemberRole',
    'WebhookType',
    'ExpireBehaviour',
    'ExpireBehavior',
    'StickerType',
    'StickerFormatType',
    'InviteTarget',
    'VideoQualityMode',
    'ComponentType',
    'ButtonStyle',
    'TextStyle',
    'PrivacyLevel',
    'InteractionType',
    'InteractionResponseType',
    'NSFWLevel',
    'MFALevel',
    'Locale',
    'EntityType',
    'EventStatus',
    'AppCommandType',
    'AppCommandOptionType',
    'AppCommandPermissionType',
    'AutoModRuleTriggerType',
    'AutoModRuleEventType',
    'AutoModRuleActionType',
    'ForumLayoutType',
    'ForumOrderType',
    'SelectDefaultValueType',
    'SKUType',
    'EntitlementType',
    'EntitlementOwnerType',
    'PollLayoutType',
    'InviteType',
    'ReactionType',
    'VoiceChannelEffectAnimationType',
    'SubscriptionStatus',
    'MessageReferenceType',
    'StatusDisplayType',
    'OnboardingPromptType',
    'OnboardingMode',
    'SeparatorSpacing',
    'MediaItemLoadingState',
    'CollectibleType',
    'NameplatePalette',
)


def _create_value_cls(name: str, comparable: bool):
    # All the type ignores here are due to the type checker being unable to recognise
    # Runtime type creation without exploding.
    cls = namedtuple('_EnumValue_' + name, 'name value')
    cls.__repr__ = lambda self: f'<{name}.{self.name}: {self.value!r}>'
    cls.__str__ = lambda self: f'{name}.{self.name}'
    if comparable:
        cls.__le__ = lambda self, other: isinstance(other, self.__class__) and self.value <= other.value
        cls.__ge__ = lambda self, other: isinstance(other, self.__class__) and self.value >= other.value
        cls.__lt__ = lambda self, other: isinstance(other, self.__class__) and self.value < other.value
        cls.__gt__ = lambda self, other: isinstance(other, self.__class__) and self.value > other.value
    return cls


def _is_descriptor(obj):
    return hasattr(obj, '__get__') or hasattr(obj, '__set__') or hasattr(obj, '__delete__')


class EnumMeta(type):
    if TYPE_CHECKING:
        __name__: ClassVar[str]
        _enum_member_names_: ClassVar[List[str]]
        _enum_member_map_: ClassVar[Dict[str, Any]]
        _enum_value_map_: ClassVar[Dict[Any, Any]]

    def __new__(
        cls,
        name: str,
        bases: Tuple[type, ...],
        attrs: Dict[str, Any],
        *,
        comparable: bool = False,
    ) -> EnumMeta:
        value_mapping = {}
        member_mapping = {}
        member_names = []

        value_cls = _create_value_cls(name, comparable)
        for key, value in list(attrs.items()):
            is_descriptor = _is_descriptor(value)
            if key[0] == '_' and not is_descriptor:
                continue

            # Special case classmethod to just pass through
            if isinstance(value, classmethod):
                continue

            if is_descriptor:
                setattr(value_cls, key, value)
                del attrs[key]
                continue

            try:
                new_value = value_mapping[value]
            except KeyError:
                new_value = value_cls(name=key, value=value)
                value_mapping[value] = new_value
                member_names.append(key)

            member_mapping[key] = new_value
            attrs[key] = new_value

        attrs['_enum_value_map_'] = value_mapping
        attrs['_enum_member_map_'] = member_mapping
        attrs['_enum_member_names_'] = member_names
        attrs['_enum_value_cls_'] = value_cls
        actual_cls = super().__new__(cls, name, bases, attrs)
        value_cls._actual_enum_cls_ = actual_cls  # type: ignore # Runtime attribute isn't understood
        return actual_cls

    def __iter__(cls) -> Iterator[Any]:
        return (cls._enum_member_map_[name] for name in cls._enum_member_names_)

    def __reversed__(cls) -> Iterator[Any]:
        return (cls._enum_member_map_[name] for name in reversed(cls._enum_member_names_))

    def __len__(cls) -> int:
        return len(cls._enum_member_names_)

    def __repr__(cls) -> str:
        return f'<enum {cls.__name__}>'

    @property
    def __members__(cls) -> Mapping[str, Any]:
        return types.MappingProxyType(cls._enum_member_map_)

    def __call__(cls, value: str) -> Any:
        try:
            return cls._enum_value_map_[value]
        except (KeyError, TypeError):
            raise ValueError(f'{value!r} is not a valid {cls.__name__}')

    def __getitem__(cls, key: str) -> Any:
        return cls._enum_member_map_[key]

    def __setattr__(cls, name: str, value: Any) -> None:
        raise TypeError('Enums are immutable.')

    def __delattr__(cls, attr: str) -> None:
        raise TypeError('Enums are immutable')

    def __instancecheck__(self, instance: Any) -> bool:
        # isinstance(x, Y)
        # -> __instancecheck__(Y, x)
        try:
            return instance._actual_enum_cls_ is self
        except AttributeError:
            return False


if TYPE_CHECKING:
    from enum import Enum
else:

    class Enum(metaclass=EnumMeta):
        @classmethod
        def try_value(cls, value):
            try:
                return cls._enum_value_map_[value]
            except (KeyError, TypeError):
                return value


class ChannelType(Enum):
    text = 0
    private = 1
    voice = 2
    group = 3
    category = 4
    news = 5
    news_thread = 10
    public_thread = 11
    private_thread = 12
    stage_voice = 13
    forum = 15
    media = 16

    def __str__(self) -> str:
        return self.name


class MessageReferenceType(Enum):
    default = 0
    reply = 0
    forward = 1


class MessageType(Enum):
    default = 0
    recipient_add = 1
    recipient_remove = 2
    call = 3
    channel_name_change = 4
    channel_icon_change = 5
    pins_add = 6
    new_member = 7
    premium_guild_subscription = 8
    premium_guild_tier_1 = 9
    premium_guild_tier_2 = 10
    premium_guild_tier_3 = 11
    channel_follow_add = 12
    guild_stream = 13
    guild_discovery_disqualified = 14
    guild_discovery_requalified = 15
    guild_discovery_grace_period_initial_warning = 16
    guild_discovery_grace_period_final_warning = 17
    thread_created = 18
    reply = 19
    chat_input_command = 20
    thread_starter_message = 21
    guild_invite_reminder = 22
    context_menu_command = 23
    auto_moderation_action = 24
    role_subscription_purchase = 25
    interaction_premium_upsell = 26
    stage_start = 27
    stage_end = 28
    stage_speaker = 29
    stage_raise_hand = 30
    stage_topic = 31
    guild_application_premium_subscription = 32
    guild_incident_alert_mode_enabled = 36
    guild_incident_alert_mode_disabled = 37
    guild_incident_report_raid = 38
    guild_incident_report_false_alarm = 39
    purchase_notification = 44
    poll_result = 46
    emoji_added = 63

    def is_deletable(self) -> bool:
        return self not in {
            MessageType.recipient_add,
            MessageType.recipient_remove,
            MessageType.call,
            MessageType.channel_name_change,
            MessageType.channel_icon_change,
            MessageType.thread_starter_message,
        }


class SpeakingState(Enum):
    none = 0
    voice = 1
    soundshare = 2
    priority = 4

    def __str__(self) -> str:
        return self.name

    def __int__(self) -> int:
        return self.value


class VerificationLevel(Enum, comparable=True):
    none = 0
    low = 1
    medium = 2
    high = 3
    highest = 4

    def __str__(self) -> str:
        return self.name


class ContentFilter(Enum, comparable=True):
    disabled = 0
    no_role = 1
    all_members = 2

    def __str__(self) -> str:
        return self.name


class Status(Enum):
    online = 'online'
    offline = 'offline'
    idle = 'idle'
    dnd = 'dnd'
    do_not_disturb = 'dnd'
    invisible = 'invisible'

    def __str__(self) -> str:
        return self.value


class DefaultAvatar(Enum):
    blurple = 0
    grey = 1
    gray = 1
    green = 2
    orange = 3
    red = 4
    pink = 5

    def __str__(self) -> str:
        return self.name


class NotificationLevel(Enum, comparable=True):
    all_messages = 0
    only_mentions = 1


class AuditLogActionCategory(Enum):
    create = 1
    delete = 2
    update = 3


class AuditLogAction(Enum):
    # fmt: off
    guild_update                                      = 1
    channel_create                                    = 10
    channel_update                                    = 11
    channel_delete                                    = 12
    overwrite_create                                  = 13
    overwrite_update                                  = 14
    overwrite_delete                                  = 15
    kick                                              = 20
    member_prune                                      = 21
    ban                                               = 22
    unban                                             = 23
    member_update                                     = 24
    member_role_update                                = 25
    member_move                                       = 26
    member_disconnect                                 = 27
    bot_add                                           = 28
    role_create                                       = 30
    role_update                                       = 31
    role_delete                                       = 32
    invite_create                                     = 40
    invite_update                                     = 41
    invite_delete                                     = 42
    webhook_create                                    = 50
    webhook_update                                    = 51
    webhook_delete                                    = 52
    emoji_create                                      = 60
    emoji_update                                      = 61
    emoji_delete                                      = 62
    message_delete                                    = 72
    message_bulk_delete                               = 73
    message_pin                                       = 74
    message_unpin                                     = 75
    integration_create                                = 80
    integration_update                                = 81
    integration_delete                                = 82
    stage_instance_create                             = 83
    stage_instance_update                             = 84
    stage_instance_delete                             = 85
    sticker_create                                    = 90
    sticker_update                                    = 91
    sticker_delete                                    = 92
    scheduled_event_create                            = 100
    scheduled_event_update                            = 101
    scheduled_event_delete                            = 102
    thread_create                                     = 110
    thread_update                                     = 111
    thread_delete                                     = 112
    app_command_permission_update                     = 121
    soundboard_sound_create                           = 130
    soundboard_sound_update                           = 131
    soundboard_sound_delete                           = 132
    automod_rule_create                               = 140
    automod_rule_update                               = 141
    automod_rule_delete                               = 142
    automod_block_message                             = 143
    automod_flag_message                              = 144
    automod_timeout_member                            = 145
    automod_quarantine_user                           = 146
    creator_monetization_request_created              = 150
    creator_monetization_terms_accepted               = 151
    onboarding_prompt_create                          = 163
    onboarding_prompt_update                          = 164
    onboarding_prompt_delete                          = 165
    onboarding_create                                 = 166
    onboarding_update                                 = 167
    home_settings_create                              = 190
    home_settings_update                              = 191
    # fmt: on

    @property
    def category(self) -> Optional[AuditLogActionCategory]:
        # fmt: off
        lookup: Dict[AuditLogAction, Optional[AuditLogActionCategory]] = {
            AuditLogAction.guild_update:                             AuditLogActionCategory.update,
            AuditLogAction.channel_create:                           AuditLogActionCategory.create,
            AuditLogAction.channel_update:                           AuditLogActionCategory.update,
            AuditLogAction.channel_delete:                           AuditLogActionCategory.delete,
            AuditLogAction.overwrite_create:                         AuditLogActionCategory.create,
            AuditLogAction.overwrite_update:                         AuditLogActionCategory.update,
            AuditLogAction.overwrite_delete:                         AuditLogActionCategory.delete,
            AuditLogAction.kick:                                     None,
            AuditLogAction.member_prune:                             None,
            AuditLogAction.ban:                                      None,
            AuditLogAction.unban:                                    None,
            AuditLogAction.member_update:                            AuditLogActionCategory.update,
            AuditLogAction.member_role_update:                       AuditLogActionCategory.update,
            AuditLogAction.member_move:                              None,
            AuditLogAction.member_disconnect:                        None,
            AuditLogAction.bot_add:                                  None,
            AuditLogAction.role_create:                              AuditLogActionCategory.create,
            AuditLogAction.role_update:                              AuditLogActionCategory.update,
            AuditLogAction.role_delete:                              AuditLogActionCategory.delete,
            AuditLogAction.invite_create:                            AuditLogActionCategory.create,
            AuditLogAction.invite_update:                            AuditLogActionCategory.update,
            AuditLogAction.invite_delete:                            AuditLogActionCategory.delete,
            AuditLogAction.webhook_create:                           AuditLogActionCategory.create,
            AuditLogAction.webhook_update:                           AuditLogActionCategory.update,
            AuditLogAction.webhook_delete:                           AuditLogActionCategory.delete,
            AuditLogAction.emoji_create:                             AuditLogActionCategory.create,
            AuditLogAction.emoji_update:                             AuditLogActionCategory.update,
            AuditLogAction.emoji_delete:                             AuditLogActionCategory.delete,
            AuditLogAction.message_delete:                           AuditLogActionCategory.delete,
            AuditLogAction.message_bulk_delete:                      AuditLogActionCategory.delete,
            AuditLogAction.message_pin:                              None,
            AuditLogAction.message_unpin:                            None,
            AuditLogAction.integration_create:                       AuditLogActionCategory.create,
            AuditLogAction.integration_update:                       AuditLogActionCategory.update,
            AuditLogAction.integration_delete:                       AuditLogActionCategory.delete,
            AuditLogAction.stage_instance_create:                    AuditLogActionCategory.create,
            AuditLogAction.stage_instance_update:                    AuditLogActionCategory.update,
            AuditLogAction.stage_instance_delete:                    AuditLogActionCategory.delete,
            AuditLogAction.sticker_create:                           AuditLogActionCategory.create,
            AuditLogAction.sticker_update:                           AuditLogActionCategory.update,
            AuditLogAction.sticker_delete:                           AuditLogActionCategory.delete,
            AuditLogAction.scheduled_event_create:                   AuditLogActionCategory.create,
            AuditLogAction.scheduled_event_update:                   AuditLogActionCategory.update,
            AuditLogAction.scheduled_event_delete:                   AuditLogActionCategory.delete,
            AuditLogAction.thread_create:                            AuditLogActionCategory.create,
            AuditLogAction.thread_delete:                            AuditLogActionCategory.delete,
            AuditLogAction.thread_update:                            AuditLogActionCategory.update,
            AuditLogAction.app_command_permission_update:            AuditLogActionCategory.update,
            AuditLogAction.automod_rule_create:                      AuditLogActionCategory.create,
            AuditLogAction.automod_rule_update:                      AuditLogActionCategory.update,
            AuditLogAction.automod_rule_delete:                      AuditLogActionCategory.delete,
            AuditLogAction.automod_block_message:                    None,
            AuditLogAction.automod_flag_message:                     None,
            AuditLogAction.automod_timeout_member:                   None,
            AuditLogAction.automod_quarantine_user:                  None,
            AuditLogAction.creator_monetization_request_created:     None,
            AuditLogAction.creator_monetization_terms_accepted:      None,
            AuditLogAction.soundboard_sound_create:                  AuditLogActionCategory.create,
            AuditLogAction.soundboard_sound_update:                  AuditLogActionCategory.update,
            AuditLogAction.soundboard_sound_delete:                  AuditLogActionCategory.delete,
            AuditLogAction.onboarding_prompt_create:                 AuditLogActionCategory.create,
            AuditLogAction.onboarding_prompt_update:                 AuditLogActionCategory.update,
            AuditLogAction.onboarding_prompt_delete:                 AuditLogActionCategory.delete,
            AuditLogAction.onboarding_create:                        AuditLogActionCategory.create,
            AuditLogAction.onboarding_update:                        AuditLogActionCategory.update,
            AuditLogAction.home_settings_create:                     AuditLogActionCategory.create,
            AuditLogAction.home_settings_update:                     AuditLogActionCategory.update,
        }
        # fmt: on
        return lookup.get(self, None)

    @property
    def target_type(self) -> Optional[str]:
        v = self.value
        if v == -1:
            return 'all'
        elif v < 10:
            return 'guild'
        elif v < 20:
            return 'channel'
        elif v < 30:
            return 'user'
        elif v < 40:
            return 'role'
        elif v < 50:
            return 'invite'
        elif v < 60:
            return 'webhook'
        elif v < 70:
            return 'emoji'
        elif v == 73:
            return 'channel'
        elif v < 80:
            return 'message'
        elif v < 83:
            return 'integration'
        elif v < 90:
            return 'stage_instance'
        elif v < 93:
            return 'sticker'
        elif v < 103:
            return 'guild_scheduled_event'
        elif v < 113:
            return 'thread'
        elif v < 122:
            return 'integration_or_app_command'
        elif 139 < v < 143:
            return 'auto_moderation'
        elif v < 147:
            return 'user'
        elif v < 152:
            return 'creator_monetization'
        elif v < 166:
            return 'onboarding_prompt'
        elif v < 168:
            return 'onboarding'
        elif v < 192:
            return 'home_settings'


class UserFlags(Enum):
    staff = 1
    partner = 2
    hypesquad = 4
    bug_hunter = 8
    mfa_sms = 16
    premium_promo_dismissed = 32
    hypesquad_bravery = 64
    hypesquad_brilliance = 128
    hypesquad_balance = 256
    early_supporter = 512
    team_user = 1024
    system = 4096
    has_unread_urgent_messages = 8192
    bug_hunter_level_2 = 16384
    verified_bot = 65536
    verified_bot_developer = 131072
    discord_certified_moderator = 262144
    bot_http_interactions = 524288
    spammer = 1048576
    active_developer = 4194304


class ActivityType(Enum):
    unknown = -1
    playing = 0
    streaming = 1
    listening = 2
    watching = 3
    custom = 4
    competing = 5

    def __int__(self) -> int:
        return self.value


class TeamMembershipState(Enum):
    invited = 1
    accepted = 2


class TeamMemberRole(Enum):
    admin = 'admin'
    developer = 'developer'
    read_only = 'read_only'


class WebhookType(Enum):
    incoming = 1
    channel_follower = 2
    application = 3


class ExpireBehaviour(Enum):
    remove_role = 0
    kick = 1


ExpireBehavior = ExpireBehaviour


class StickerType(Enum):
    standard = 1
    guild = 2


class StickerFormatType(Enum):
    png = 1
    apng = 2
    lottie = 3
    gif = 4

    @property
    def file_extension(self) -> str:
        # fmt: off
        lookup: Dict[StickerFormatType, str] = {
            StickerFormatType.png: 'png',
            StickerFormatType.apng: 'png',
            StickerFormatType.lottie: 'json',
            StickerFormatType.gif: 'gif',
        }
        # fmt: on
        return lookup.get(self, 'png')


class InviteTarget(Enum):
    unknown = 0
    stream = 1
    embedded_application = 2


class InteractionType(Enum):
    ping = 1
    application_command = 2
    component = 3
    autocomplete = 4
    modal_submit = 5


class InteractionResponseType(Enum):
    pong = 1
    # ack = 2 (deprecated)
    # channel_message = 3 (deprecated)
    channel_message = 4  # (with source)
    deferred_channel_message = 5  # (with source)
    deferred_message_update = 6  # for components
    message_update = 7  # for components
    autocomplete_result = 8
    modal = 9  # for modals
    # premium_required = 10 (deprecated)
    launch_activity = 12


class VideoQualityMode(Enum):
    auto = 1
    full = 2

    def __int__(self) -> int:
        return self.value


class ComponentType(Enum):
    action_row = 1
    button = 2
    select = 3
    string_select = 3
    text_input = 4
    user_select = 5
    role_select = 6
    mentionable_select = 7
    channel_select = 8
    section = 9
    text_display = 10
    thumbnail = 11
    media_gallery = 12
    file = 13
    separator = 14
    container = 17
    label = 18
    file_upload = 19
    # checkpoint = 20
    radio_group = 21
    checkbox_group = 22
    checkbox = 23

    def __int__(self) -> int:
        return self.value


class ButtonStyle(Enum):
    primary = 1
    secondary = 2
    success = 3
    danger = 4
    link = 5
    premium = 6

    # Aliases
    blurple = 1
    grey = 2
    gray = 2
    green = 3
    red = 4
    url = 5

    def __int__(self) -> int:
        return self.value


class TextStyle(Enum):
    short = 1
    paragraph = 2

    # Aliases
    long = 2

    def __int__(self) -> int:
        return self.value


class PrivacyLevel(Enum):
    guild_only = 2


class NSFWLevel(Enum, comparable=True):
    default = 0
    explicit = 1
    safe = 2
    age_restricted = 3


class MFALevel(Enum, comparable=True):
    disabled = 0
    require_2fa = 1


_UNICODE_LANG_MAP: Dict[str, str] = {
    'bg': 'bg-BG',
    'zh-CN': 'zh-CN',
    'zh-TW': 'zh-TW',
    'hr': 'hr-HR',
    'cs': 'cs-CZ',
    'da': 'da-DK',
    'nl': 'nl-NL',
    'en-US': 'en-US',
    'en-GB': 'en-GB',
    'fi': 'fi-FI',
    'fr': 'fr-FR',
    'de': 'de-DE',
    'el': 'el-GR',
    'hi': 'hi-IN',
    'hu': 'hu-HU',
    'id': 'id-ID',
    'it': 'it-IT',
    'ja': 'ja-JP',
    'ko': 'ko-KR',
    'lt': 'lt-LT',
    'no': 'no-NO',
    'pl': 'pl-PL',
    'pt-BR': 'pt-BR',
    'ro': 'ro-RO',
    'ru': 'ru-RU',
    'es-ES': 'es-ES',
    'es-419': 'es-419',
    'sv-SE': 'sv-SE',
    'th': 'th-TH',
    'tr': 'tr-TR',
    'uk': 'uk-UA',
    'vi': 'vi-VN',
}


class Locale(Enum):
    american_english = 'en-US'
    british_english = 'en-GB'
    bulgarian = 'bg'
    chinese = 'zh-CN'
    taiwan_chinese = 'zh-TW'
    croatian = 'hr'
    czech = 'cs'
    indonesian = 'id'
    danish = 'da'
    dutch = 'nl'
    finnish = 'fi'
    french = 'fr'
    german = 'de'
    greek = 'el'
    hindi = 'hi'
    hungarian = 'hu'
    italian = 'it'
    japanese = 'ja'
    korean = 'ko'
    latin_american_spanish = 'es-419'
    lithuanian = 'lt'
    norwegian = 'no'
    polish = 'pl'
    brazil_portuguese = 'pt-BR'
    romanian = 'ro'
    russian = 'ru'
    spain_spanish = 'es-ES'
    swedish = 'sv-SE'
    thai = 'th'
    turkish = 'tr'
    ukrainian = 'uk'
    vietnamese = 'vi'

    def __str__(self) -> str:
        return self.value

    @property
    def language_code(self) -> str:
        return _UNICODE_LANG_MAP.get(self.value, self.value)


E = TypeVar('E', bound='Enum')


class EntityType(Enum):
    stage_instance = 1
    voice = 2
    external = 3


class EventStatus(Enum):
    scheduled = 1
    active = 2
    completed = 3
    canceled = 4

    ended = 3
    cancelled = 4


class AppCommandOptionType(Enum):
    subcommand = 1
    subcommand_group = 2
    string = 3
    integer = 4
    boolean = 5
    user = 6
    channel = 7
    role = 8
    mentionable = 9
    number = 10
    attachment = 11


class AppCommandType(Enum):
    chat_input = 1
    user = 2
    message = 3


class AppCommandPermissionType(Enum):
    role = 1
    user = 2
    channel = 3


class AutoModRuleTriggerType(Enum):
    keyword = 1
    harmful_link = 2
    spam = 3
    keyword_preset = 4
    mention_spam = 5
    member_profile = 6


class AutoModRuleEventType(Enum):
    message_send = 1
    member_update = 2


class AutoModRuleActionType(Enum):
    block_message = 1
    send_alert_message = 2
    timeout = 3
    block_member_interactions = 4


class ForumLayoutType(Enum):
    not_set = 0
    list_view = 1
    gallery_view = 2


class ForumOrderType(Enum):
    latest_activity = 0
    creation_date = 1


class SelectDefaultValueType(Enum):
    user = 'user'
    role = 'role'
    channel = 'channel'


class SKUType(Enum):
    durable = 2
    consumable = 3
    subscription = 5
    subscription_group = 6


class EntitlementType(Enum):
    purchase = 1
    premium_subscription = 2
    developer_gift = 3
    test_mode_purchase = 4
    free_purchase = 5
    user_gift = 6
    premium_purchase = 7
    application_subscription = 8


class EntitlementOwnerType(Enum):
    guild = 1
    user = 2


class PollLayoutType(Enum):
    default = 1


class InviteType(Enum):
    guild = 0
    group_dm = 1
    friend = 2


class ReactionType(Enum):
    normal = 0
    burst = 1


class VoiceChannelEffectAnimationType(Enum):
    premium = 0
    basic = 1


class SubscriptionStatus(Enum):
    active = 0
    ending = 1
    inactive = 2


class StatusDisplayType(Enum):
    name = 0  # pyright: ignore[reportAssignmentType]
    state = 1
    details = 2


class OnboardingPromptType(Enum):
    multiple_choice = 0
    dropdown = 1


class OnboardingMode(Enum):
    default = 0
    advanced = 1


class SeparatorSpacing(Enum):
    small = 1
    large = 2


class MediaItemLoadingState(Enum):
    unknown = 0
    loading = 1
    loaded = 2
    not_found = 3


class CollectibleType(Enum):
    nameplate = 'nameplate'


class NameplatePalette(Enum):
    crimson = 'crimson'
    berry = 'berry'
    sky = 'sky'
    teal = 'teal'
    forest = 'forest'
    bubble_gum = 'bubble_gum'
    violet = 'violet'
    cobalt = 'cobalt'
    clover = 'clover'
    lemon = 'lemon'
    white = 'white'


def create_unknown_value(cls: Type[E], val: Any) -> E:
    value_cls = cls._enum_value_cls_  # type: ignore # This is narrowed below
    name = f'unknown_{val}'
    return value_cls(name=name, value=val)


def try_enum(cls: Type[E], val: Any) -> E:
    """A function that tries to turn the value into enum ``cls``.

    If it fails it returns a proxy invalid value instead.
    """

    try:
        return cls._enum_value_map_[val]  # type: ignore # All errors are caught below
    except (KeyError, TypeError, AttributeError):
        return create_unknown_value(cls, val)



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations
from typing import Dict, Iterator, List, Literal, Optional, Tuple

from .state import ConnectionState

__all__ = (
    'SearchAuthorType',
    'SearchHas',
    'SearchEmbedType',
    'SearchResult',
)


class _BaseSearchType:
    __slots__ = ('value',)

    def __init__(self) -> None:
        self.value: Dict[str, bool] = {}

    def __iter__(self) -> Iterator[str]:
        for key, value in self.value.items():
            if value:
                yield key
            else:
                yield f'-{key}'


class SearchAuthorType(_BaseSearchType):
    """Represents the search ``author_type`` filter.

    .. versionadded:: 2.7

    Parameters
    ----------
    user: Optional[:class:`bool`]
        Whether to include regular users in the search. This can be ``False`` to exclude instead.
    bot: Optional[:class:`bool`]
        Whether to include bots in the search. This can be ``False`` to exclude instead.
    webhook: Optional[:class:`bool`]
        Whether to include webhooks in the search. This can be ``False`` to exclude instead.
    """

    __slots__ = ()

    def __init__(self, *, user: Optional[bool] = None, bot: Optional[bool] = None, webhook: Optional[bool] = None) -> None:
        super().__init__()

        if user is not None:
            self.value['user'] = user
        if bot is not None:
            self.value['bot'] = bot
        if webhook is not None:
            self.value['webhook'] = webhook

    @property
    def user(self) -> Optional[bool]:
        """Optional[bool]: Whether regular users are included in the search."""
        return self.value.get('user', None)

    @user.setter
    def user(self, value: Optional[bool]) -> None:
        if value is None:
            self.value.pop('user', None)
        else:
            self.value['user'] = value

    @property
    def bot(self) -> Optional[bool]:
        """Optional[bool]: Whether bots are included in the search."""
        return self.value.get('bot', None)

    @bot.setter
    def bot(self, value: Optional[bool]) -> None:
        if value is None:
            self.value.pop('bot', None)
        else:
            self.value['bot'] = value

    @property
    def webhook(self) -> Optional[bool]:
        """Optional[bool]: Whether webhooks are included in the search."""
        return self.value.get('webhook', None)

    @webhook.setter
    def webhook(self, value: Optional[bool]) -> None:
        if value is None:
            self.value.pop('webhook', None)
        else:
            self.value['webhook'] = value


class SearchHas(_BaseSearchType):
    """Represents the search ``has`` filter.

    .. versionadded:: 2.7

    Parameters
    ----------
    link: Optional[:class:`bool`]
        Whether to include messages with a link. This can be ``False`` to exclude instead.
    embed: Optional[:class:`bool`]
        Whether to include messages with an embed. This can be ``False`` to exclude instead.
    file: Optional[:class:`bool`]
        Whether to include messages with an uploaded file. This can be ``False`` to exclude instead.
    image: Optional[:class:`bool`]
        Whether to include messages with an image. This can be ``False`` to exclude instead.
    video: Optional[:class:`bool`]
        Whether to include messages with a video. This can be ``False`` to exclude instead.
    sound: Optional[:class:`bool`]
        Whether to include messages with an uploaded audio file. This can be ``False`` to exclude instead.
    sticker: Optional[:class:`bool`]
        Whether to include messages with a sticker. This can be ``False`` to exclude instead.
    poll: Optional[:class:`bool`]
        Whether to include messages with a poll. This can be ``False`` to exclude instead.
    snapshot: Optional[:class:`bool`]
        Whether to include messages with a forwarded message. This can be ``False`` to exclude instead.
    """

    def __init__(
        self,
        *,
        link: Optional[bool] = None,
        embed: Optional[bool] = None,
        file: Optional[bool] = None,
        image: Optional[bool] = None,
        video: Optional[bool] = None,
        sound: Optional[bool] = None,
        sticker: Optional[bool] = None,
        poll: Optional[bool] = None,
        snapshot: Optional[bool] = None,
    ) -> None:
        super().__init__()
        if link is not None:
            self.value['link'] = link
        if embed is not None:
            self.value['embed'] = embed
        if file is not None:
            self.value['file'] = file
        if image is not None:
            self.value['image'] = image
        if video is not None:
            self.value['video'] = video
        if sound is not None:
            self.value['sound'] = sound
        if sticker is not None:
            self.value['sticker'] = sticker
        if poll is not None:
            self.value['poll'] = poll
        if snapshot is not None:
            self.value['snapshot'] = snapshot

    @property
    def link(self) -> Optional[bool]:
        """Optional[:class:`bool`]: Whether to include messages with a link."""
        return self.value.get('link', None)

    @link.setter
    def link(self, value: Optional[bool]) -> None:
        if value is None:
            self.value.pop('link', None)
        else:
            self.value['link'] = value

    @property
    def embed(self) -> Optional[bool]:
        """Optional[:class:`bool`]: Whether to include messages with an embed."""
        return self.value.get('embed', None)

    @embed.setter
    def embed(self, value: Optional[bool]) -> None:
        if value is None:
            self.value.pop('embed', None)
        else:
            self.value['embed'] = value

    @property
    def file(self) -> Optional[bool]:
        """Optional[:class:`bool`]: Whether to include messages with an uploaded file."""
        return self.value.get('file', None)

    @file.setter
    def file(self, value: Optional[bool]) -> None:
        if value is None:
            self.value.pop('file', None)
        else:
            self.value['file'] = value

    @property
    def image(self) -> Optional[bool]:
        """Optional[:class:`bool`]: Whether to include messages with an image."""
        return self.value.get('image', None)

    @image.setter
    def image(self, value: Optional[bool]) -> None:
        if value is None:
            self.value.pop('image', None)
        else:
            self.value['image'] = value

    @property
    def video(self) -> Optional[bool]:
        """Optional[:class:`bool`]: Whether to include messages with a video."""
        return self.value.get('video', None)

    @video.setter
    def video(self, value: Optional[bool]) -> None:
        if value is None:
            self.value.pop('video', None)
        else:
            self.value['video'] = value

    @property
    def sound(self) -> Optional[bool]:
        """Optional[:class:`bool`]: Whether to include messages with an uploaded audio file."""
        return self.value.get('sound', None)

    @sound.setter
    def sound(self, value: Optional[bool]) -> None:
        if value is None:
            self.value.pop('sound', None)
        else:
            self.value['sound'] = value

    @property
    def sticker(self) -> Optional[bool]:
        """Optional[:class:`bool`]: Whether to include messages with a sticker."""
        return self.value.get('sticker', None)

    @sticker.setter
    def sticker(self, value: Optional[bool]) -> None:
        if value is None:
            self.value.pop('sticker', None)
        else:
            self.value['sticker'] = value

    @property
    def poll(self) -> Optional[bool]:
        """Optional[:class:`bool`]: Whether to include messages with a poll."""
        return self.value.get('poll', None)

    @poll.setter
    def poll(self, value: Optional[bool]) -> None:
        if value is None:
            self.value.pop('poll', None)
        else:
            self.value['poll'] = value

    @property
    def snapshot(self) -> Optional[bool]:
        """Optional[:class:`bool`]: Whether to include messages with a forwarded message."""
        return self.value.get('snapshot', None)

    @snapshot.setter
    def snapshot(self, value: Optional[bool]) -> None:
        if value is None:
            self.value.pop('snapshot', None)
        else:
            self.value['snapshot'] = value


class SearchEmbedType(_BaseSearchType):
    """Represents the search ``embed_type`` filter.

    .. versionadded:: 2.7

    Parameters
    ----------
    image: Optional[:class:`bool`]
        Whether to include image embeds. This *cannot* be ``False`` to exclude.
    video: Optional[:class:`bool`]
        Whether to include video embeds. This *cannot* be ``False`` to exclude.
    gif: Optional[:class:`bool`]
        Whether to include gif embeds. This *cannot* be ``False`` to exclude.
    sound: Optional[:class:`bool`]
        Whether to include sound embeds. This *cannot* be ``False`` to exclude.
    article: Optional[:class:`bool`]
        Whether to include article embeds. This *cannot* be ``False`` to exclude.
    """

    def __init__(
        self,
        *,
        image: Literal[True, None] = None,
        video: Literal[True, None] = None,
        gif: Literal[True, None] = None,
        sound: Literal[True, None] = None,
        article: Literal[True, None] = None,
    ) -> None:
        super().__init__()

        if image:
            self.value['image'] = True
        if video:
            self.value['video'] = True
        if gif:
            self.value['gif'] = True
        if sound:
            self.value['sound'] = True
        if article:
            self.value['article'] = True

    @property
    def image(self) -> Optional[bool]:
        """Optional[:class:`bool`]: Whether image embeds are included in the search."""
        return self.value.get('image', None)

    @image.setter
    def image(self, value: Literal[True, None]) -> None:
        if value:
            self.value['image'] = True
        else:
            self.value.pop('image', None)

    @property
    def video(self) -> Optional[bool]:
        """Optional[:class:`bool`]: Whether video embeds are included in the search."""
        return self.value.get('video', None)

    @video.setter
    def video(self, value: Literal[True, None]) -> None:
        if value:
            self.value['video'] = True
        else:
            self.value.pop('video', None)

    @property
    def gif(self) -> Optional[bool]:
        """Optional[:class:`bool`]: Whether gif embeds are included in the search."""
        return self.value.get('gif', None)

    @gif.setter
    def gif(self, value: Literal[True, None]) -> None:
        if value:
            self.value['gif'] = True
        else:
            self.value.pop('gif', None)

    @property
    def sound(self) -> Optional[bool]:
        """Optional[:class:`bool`]: Whether sound embeds are included in the search."""
        return self.value.get('sound', None)

    @sound.setter
    def sound(self, value: Literal[True, None]) -> None:
        if value:
            self.value['sound'] = True
        else:
            self.value.pop('sound', None)

    @property
    def article(self) -> Optional[bool]:
        """Optional[:class:`bool`]: Whether article embeds are included in the search."""
        return self.value.get('article', None)

    @article.setter
    def article(self, value: Literal[True, None]) -> None:
        if value:
            self.value['article'] = True
        else:
            self.value.pop('article', None)


# class SearchResult:
#     """Represents the search results from :meth:`Guild.search`.

#     .. versionadded:: 2.7

#     Attributes
#     -----------
#     total_results: :class:`int`
#         The total number of results that match the search criteria.
#     """

#     def __init__(self, total_results: int, results: list[discord.Message]) -> None:
#         self.total_results = total_results
#         self.results = results

#     def __repr__(self) -> str:
#         return f'<SearchResult total_results={self.total_results} results={len(self.results)}>'


class _SearchResultFuture:
    def __init__(
        self,
        *,
        state: ConnectionState,
        params: List[Tuple[str, str]],
        limit: Optional[int],
        wait_for_index: bool,
    ) -> None:
        self.params: List[Tuple[str, str]] = params
        self.state: ConnectionState = state
        self.limit: Optional[int] = limit
        self.wait_for_index: bool = wait_for_index



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations
from typing import TYPE_CHECKING, Any, List, Literal, Optional, Tuple, TypeVar, Dict

import os

from ..utils import MISSING
from ..components import CheckboxGroupComponent, CheckboxComponent, CheckboxGroupOption
from ..enums import ComponentType
from .item import Item

if TYPE_CHECKING:
    from typing_extensions import Self

    from ..interactions import Interaction
    from ..types.interactions import (
        ModalSubmitCheckboxGroupInteractionData as ModalSubmitCheckboxGroupInteractionDataPayload,
        ModalSubmitCheckboxInteractionData as ModalSubmitCheckboxInteractionDataPayload,
    )
    from ..types.components import (
        CheckboxGroupComponent as CheckboxGroupComponentPayload,
        CheckboxComponent as CheckboxComponentPayload,
    )
    from .view import BaseView
    from ..app_commands.namespace import ResolveKey


# fmt: off
__all__ = (
    'CheckboxGroup',
    'Checkbox',
)
# fmt: on

V = TypeVar('V', bound='BaseView', covariant=True)


class CheckboxGroup(Item[V]):
    """Represents a checkbox group component within a modal.

    .. versionadded:: 2.7

    Parameters
    ------------
    id: Optional[:class:`int`]
        The ID of the component. This must be unique across the view.
    custom_id: Optional[:class:`str`]
        The custom ID of the component.
    options: List[:class:`discord.CheckboxGroupOption`]
        A list of options that can be selected in this checkbox group.
        Can only contain up to 10 items.
    max_values: Optional[:class:`int`]
        The maximum number of options that can be selected in this component.
        Must be between 1 and 10. Defaults to 1.
    min_values: Optional[:class:`int`]
        The minimum number of options that must be selected in this component.
        Must be between 0 and 10. Defaults to 0.
    required: :class:`bool`
        Whether this component is required to be filled before submitting the modal.
        Defaults to ``True``.
    """

    __item_repr_attributes__: Tuple[str, ...] = (
        'id',
        'custom_id',
        'options',
        'required',
    )

    def __init__(
        self,
        *,
        custom_id: str = MISSING,
        required: bool = True,
        min_values: Optional[int] = None,
        max_values: Optional[int] = None,
        options: List[CheckboxGroupOption] = MISSING,
        id: Optional[int] = None,
    ) -> None:
        super().__init__()
        self._provided_custom_id = custom_id is not MISSING
        custom_id = os.urandom(16).hex() if custom_id is MISSING else custom_id
        if not isinstance(custom_id, str):
            raise TypeError(f'expected custom_id to be str not {custom_id.__class__.__name__}')

        self._underlying: CheckboxGroupComponent = CheckboxGroupComponent._raw_construct(
            id=id,
            custom_id=custom_id,
            required=required,
            options=options or [],
            min_values=min_values,
            max_values=max_values,
        )
        self.id = id
        self._values: List[str] = []

    @property
    def id(self) -> Optional[int]:
        """Optional[:class:`int`]: The ID of this component."""
        return self._underlying.id

    @id.setter
    def id(self, value: Optional[int]) -> None:
        self._underlying.id = value

    @property
    def values(self) -> List[str]:
        """List[:class:`str`]: A list of values that have been selected by the user."""
        return self._values

    @property
    def custom_id(self) -> str:
        """:class:`str`: The ID of the component that gets received during an interaction."""
        return self._underlying.custom_id

    @custom_id.setter
    def custom_id(self, value: str) -> None:
        if not isinstance(value, str):
            raise TypeError('custom_id must be a str')

        self._underlying.custom_id = value
        self._provided_custom_id = True

    @property
    def type(self) -> Literal[ComponentType.checkbox_group]:
        """:class:`.ComponentType`: The type of this component."""
        return ComponentType.checkbox_group

    @property
    def options(self) -> List[CheckboxGroupOption]:
        """List[:class:`discord.CheckboxGroupOption`]: A list of options that can be selected in this menu."""
        return self._underlying.options

    @options.setter
    def options(self, value: List[CheckboxGroupOption]) -> None:
        if not isinstance(value, list) or not all(isinstance(obj, CheckboxGroupOption) for obj in value):
            raise TypeError('options must be a list of CheckboxGroupOption')
        self._underlying.options = value

    @property
    def min_values(self) -> int:
        """:class:`int`: The minimum number of options that must be selected before submitting the modal."""
        return self._underlying.min_values

    @min_values.setter
    def min_values(self, value: int) -> None:
        self._underlying.min_values = int(value)

    @property
    def max_values(self) -> int:
        """:class:`int`: The maximum number of options that can be selected before submitting the modal."""
        return self._underlying.max_values

    @max_values.setter
    def max_values(self, value: int) -> None:
        self._underlying.max_values = int(value)

    def add_option(
        self,
        *,
        label: str,
        value: str = MISSING,
        description: Optional[str] = None,
        default: bool = False,
    ) -> None:
        """Adds an option to the checkbox group.

        To append a pre-existing :class:`discord.CheckboxGroupOption` use the
        :meth:`append_option` method instead.

        Parameters
        -----------
        label: :class:`str`
            The label of the option. This is displayed to users.
            Can only be up to 100 characters.
        value: :class:`str`
            The value of the option. This is not displayed to users.
            If not given, defaults to the label.
            Can only be up to 100 characters.
        description: Optional[:class:`str`]
            An additional description of the option, if any.
            Can only be up to 100 characters.
        default: :class:`bool`
            Whether this option is selected by default.

        Raises
        -------
        ValueError
            The number of options exceeds 10.
        """

        option = CheckboxGroupOption(
            label=label,
            value=value,
            description=description,
            default=default,
        )

        self.append_option(option)

    def append_option(self, option: CheckboxGroupOption) -> None:
        """Appends an option to the checkbox group.

        Parameters
        -----------
        option: :class:`discord.CheckboxGroupOption`
            The option to append to the checkbox group.

        Raises
        -------
        ValueError
            The number of options exceeds 10.
        """

        if len(self._underlying.options) >= 10:
            raise ValueError('maximum number of options already provided (10)')

        self._underlying.options.append(option)

    @property
    def required(self) -> bool:
        """:class:`bool`: Whether the component is required or not."""
        return self._underlying.required

    @required.setter
    def required(self, value: bool) -> None:
        self._underlying.required = bool(value)

    @property
    def width(self) -> int:
        return 5

    def to_component_dict(self) -> CheckboxGroupComponentPayload:
        return self._underlying.to_dict()

    def _refresh_component(self, component: CheckboxGroupComponent) -> None:
        self._underlying = component

    def _handle_submit(
        self, interaction: Interaction, data: ModalSubmitCheckboxGroupInteractionDataPayload, resolved: Dict[ResolveKey, Any]
    ) -> None:
        self._values = data.get('values', [])

    @classmethod
    def from_component(cls, component: CheckboxGroupComponent) -> Self:
        self = cls(
            id=component.id,
            custom_id=component.custom_id,
            options=component.options,
            required=component.required,
            min_values=component.min_values,
            max_values=component.max_values,
        )
        return self

    def is_dispatchable(self) -> bool:
        return False


class Checkbox(Item[V]):
    """Represents a checkbox component within a modal.

    .. versionadded:: 2.7

    Parameters
    ------------
    id: Optional[:class:`int`]
        The ID of the component. This must be unique across the view.
    custom_id: Optional[:class:`str`]
        The custom ID of the component.
    default: :class:`bool`
        Whether this checkbox is selected by default.
    """

    __item_repr_attributes__: Tuple[str, ...] = (
        'id',
        'custom_id',
        'default',
    )

    def __init__(
        self,
        *,
        custom_id: str = MISSING,
        default: bool = False,
        id: Optional[int] = None,
    ) -> None:
        super().__init__()
        self._provided_custom_id = custom_id is not MISSING
        custom_id = os.urandom(16).hex() if custom_id is MISSING else custom_id
        if not isinstance(custom_id, str):
            raise TypeError(f'expected custom_id to be str not {custom_id.__class__.__name__}')

        self._underlying: CheckboxComponent = CheckboxComponent._raw_construct(
            id=id,
            custom_id=custom_id,
            default=default,
        )
        self.id = id
        self._value: bool = default

    @property
    def id(self) -> Optional[int]:
        """Optional[:class:`int`]: The ID of this component."""
        return self._underlying.id

    @id.setter
    def id(self, value: Optional[int]) -> None:
        self._underlying.id = value

    @property
    def value(self) -> bool:
        """:class:`bool`: ``True`` if this checkbox was selected, otherwise ``False``."""
        return self._value

    @property
    def custom_id(self) -> str:
        """:class:`str`: The ID of the component that gets received during an interaction."""
        return self._underlying.custom_id

    @custom_id.setter
    def custom_id(self, value: str) -> None:
        if not isinstance(value, str):
            raise TypeError('custom_id must be a str')

        self._underlying.custom_id = value
        self._provided_custom_id = True

    @property
    def type(self) -> Literal[ComponentType.checkbox]:
        """:class:`.ComponentType`: The type of this component."""
        return ComponentType.checkbox

    @property
    def default(self) -> bool:
        """:class:`bool`: Whether this checkbox is selected by default."""
        return self._underlying.default

    @default.setter
    def default(self, value: bool) -> None:
        self._underlying.default = bool(value)

    @property
    def width(self) -> int:
        return 5

    def to_component_dict(self) -> CheckboxComponentPayload:
        return self._underlying.to_dict()

    def _refresh_component(self, component: CheckboxComponent) -> None:
        self._underlying = component

    def _handle_submit(
        self, interaction: Interaction, data: ModalSubmitCheckboxInteractionDataPayload, resolved: Dict[ResolveKey, Any]
    ) -> None:
        self._value = data.get('value', False)

    @classmethod
    def from_component(cls, component: CheckboxComponent) -> Self:
        self = cls(
            id=component.id,
            custom_id=component.custom_id,
            default=component.default,
        )
        return self

    def is_dispatchable(self) -> bool:
        return False



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generator,
    Iterator,
    List,
    Optional,
    Sequence,
    TYPE_CHECKING,
    Set,
    Tuple,
    Type,
    Union,
)
from functools import partial
from itertools import groupby

import asyncio
import logging
import sys
import time
import os

from .item import Item, ItemCallbackType, _ItemCallback
from .select import Select
from .dynamic import DynamicItem
from ..components import (
    Component,
    ActionRow as ActionRowComponent,
    _component_factory,
    Button as ButtonComponent,
    SelectMenu as SelectComponent,
    SectionComponent,
    TextDisplay as TextDisplayComponent,
    MediaGalleryComponent,
    FileComponent,
    SeparatorComponent,
    ThumbnailComponent,
    Container as ContainerComponent,
    LabelComponent,
)
from ..utils import get as _utils_get, find as _utils_find

# fmt: off
__all__ = (
    'View',
    'LayoutView',
)
# fmt: on


if TYPE_CHECKING:
    from typing_extensions import Self
    import re

    from ..interactions import Interaction
    from .._types import ClientT
    from ..message import Message
    from ..types.components import ComponentBase as ComponentBasePayload
    from ..types.interactions import (
        ModalSubmitComponentInteractionData as ModalSubmitComponentInteractionDataPayload,
        ResolvedData as ResolvedDataPayload,
    )
    from ..state import ConnectionState
    from .modal import Modal

    ItemLike = Union[ItemCallbackType[Any, Any], Item[Any]]


_log = logging.getLogger(__name__)


def _walk_all_components(components: List[Component]) -> Iterator[Component]:
    for item in components:
        if isinstance(item, ActionRowComponent):
            yield from item.children
        elif isinstance(item, ContainerComponent):
            yield from _walk_all_components(item.children)
        elif isinstance(item, SectionComponent):
            yield from item.children
            yield item.accessory
        else:
            yield item


def _component_to_item(component: Component, parent: Optional[Item] = None) -> Item:
    if isinstance(component, ActionRowComponent):
        from .action_row import ActionRow

        item = ActionRow.from_component(component)
    elif isinstance(component, ButtonComponent):
        from .button import Button

        item = Button.from_component(component)
    elif isinstance(component, SelectComponent):
        from .select import BaseSelect

        item = BaseSelect.from_component(component)
    elif isinstance(component, SectionComponent):
        from .section import Section

        item = Section.from_component(component)
    elif isinstance(component, TextDisplayComponent):
        from .text_display import TextDisplay

        item = TextDisplay.from_component(component)
    elif isinstance(component, MediaGalleryComponent):
        from .media_gallery import MediaGallery

        item = MediaGallery.from_component(component)
    elif isinstance(component, FileComponent):
        from .file import File

        item = File.from_component(component)
    elif isinstance(component, SeparatorComponent):
        from .separator import Separator

        item = Separator.from_component(component)
    elif isinstance(component, ThumbnailComponent):
        from .thumbnail import Thumbnail

        item = Thumbnail.from_component(component)
    elif isinstance(component, ContainerComponent):
        from .container import Container

        item = Container.from_component(component)
    elif isinstance(component, LabelComponent):
        from .label import Label

        item = Label.from_component(component)
    else:
        item = Item.from_component(component)

    item._parent = parent
    return item


class _ViewWeights:
    # fmt: off
    __slots__ = (
        'weights',
    )
    # fmt: on

    def __init__(self, children: List[Item]):
        self.weights: List[int] = [0, 0, 0, 0, 0]

        key = lambda i: sys.maxsize if i.row is None else i.row
        children = sorted(children, key=key)
        for row, group in groupby(children, key=key):
            for item in group:
                self.add_item(item)

    def find_open_space(self, item: Item) -> int:
        for index, weight in enumerate(self.weights):
            if weight + item.width <= 5:
                return index

        raise ValueError('could not find open space for item')

    def add_item(self, item: Item) -> None:
        if item.row is not None:
            total = self.weights[item.row] + item.width
            if total > 5:
                raise ValueError(f'item would not fit at row {item.row} ({total} > 5 width)')
            self.weights[item.row] = total
            item._rendered_row = item.row
        else:
            index = self.find_open_space(item)
            self.weights[index] += item.width
            item._rendered_row = index

    def remove_item(self, item: Item) -> None:
        if item._rendered_row is not None:
            self.weights[item._rendered_row] -= item.width
            item._rendered_row = None

    def clear(self) -> None:
        self.weights = [0, 0, 0, 0, 0]


class BaseView:
    __discord_ui_view__: ClassVar[bool] = False
    __discord_ui_modal__: ClassVar[bool] = False
    __view_children_items__: ClassVar[Dict[str, ItemLike]] = {}

    def __init__(self, *, timeout: Optional[float] = 180.0) -> None:
        self.__timeout = timeout
        self._children: List[Item[Self]] = self._init_children()
        self.id: str = os.urandom(16).hex()
        self._cache_key: Optional[int] = None
        self.__cancel_callback: Optional[Callable[[BaseView], None]] = None
        self.__timeout_expiry: Optional[float] = None
        self.__timeout_task: Optional[asyncio.Task[None]] = None

        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            self.__stopped: Optional[asyncio.Future[bool]] = None
        else:
            self.__stopped: Optional[asyncio.Future[bool]] = loop.create_future()

        self._total_children: int = len(tuple(self.walk_children()))

    def _is_layout(self) -> bool:
        return False

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} timeout={self.timeout} children={len(self._children)}>'

    def _init_children(self) -> List[Item[Self]]:
        children = []
        parents = {}

        for name, raw in self.__view_children_items__.items():
            if isinstance(raw, Item):
                item = raw.copy()
                setattr(self, name, item)
                item._update_view(self)
                parent = getattr(item, '__discord_ui_parent__', None)
                if parent and parent._view is None:
                    parent._update_view(self)
                children.append(item)
                parents[raw] = item
            else:
                item: Item = raw.__discord_ui_model_type__(**raw.__discord_ui_model_kwargs__)
                item.callback = _ItemCallback(raw, self, item)  # type: ignore
                item._update_view(self)
                if isinstance(item, Select):
                    item.options = [option.copy() for option in item.options]
                setattr(self, raw.__name__, item)
                parent = getattr(raw, '__discord_ui_parent__', None)
                if parent:
                    parents.get(parent, parent)._children.append(item)
                    continue
                children.append(item)

        return children

    async def __timeout_task_impl(self) -> None:
        while True:
            # Guard just in case someone changes the value of the timeout at runtime
            if self.timeout is None:
                return

            if self.__timeout_expiry is None:
                return self._dispatch_timeout()

            # Check if we've elapsed our currently set timeout
            now = time.monotonic()
            if now >= self.__timeout_expiry:
                return self._dispatch_timeout()

            # Wait N seconds to see if timeout data has been refreshed
            await asyncio.sleep(self.__timeout_expiry - now)

    def is_dispatchable(self) -> bool:
        # checks whether any interactable items (buttons or selects) are present
        # in this view, and check whether this requires a state attached in case
        # of webhooks and if the view should be stored in the view store
        return any(item.is_dispatchable() for item in self.walk_children())

    def has_components_v2(self) -> bool:
        return any(c._is_v2() for c in self.children)

    def to_components(self) -> List[Dict[str, Any]]:
        return NotImplemented

    def _refresh_timeout(self) -> None:
        if self.__timeout:
            self.__timeout_expiry = time.monotonic() + self.__timeout

    def _swap_item(self, base: Item, new: DynamicItem, custom_id: str) -> None:
        # if an error is raised it is catched by the try/except block that calls
        # this function
        child_index = self._children.index(base)
        self._children[child_index] = new  # type: ignore

    @property
    def timeout(self) -> Optional[float]:
        """Optional[:class:`float`]: The timeout in seconds from last interaction with the UI before no longer accepting input.
        If ``None`` then there is no timeout.
        """
        return self.__timeout

    @timeout.setter
    def timeout(self, value: Optional[float]) -> None:
        # If the timeout task is already running this allows it to update
        # the expiry while it's running
        if self.__timeout_task is not None:
            if value is not None:
                self.__timeout_expiry = time.monotonic() + value
            else:
                self.__timeout_expiry = None

        self.__timeout = value

    def _add_count(self, value: int) -> None:
        self._total_children = max(0, self._total_children + value)

    @property
    def children(self) -> List[Item[Self]]:
        """List[:class:`Item`]: The list of children attached to this view."""
        return self._children.copy()

    @property
    def total_children_count(self) -> int:
        """:class:`int`: The total number of children in this view, including those from nested items.

        .. versionadded:: 2.6"""
        return self._total_children

    @classmethod
    def from_message(cls, message: Message, /, *, timeout: Optional[float] = 180.0) -> Union[View, LayoutView]:
        """Converts a message's components into a :class:`View`
        or :class:`LayoutView`.

        The :attr:`.Message.components` of a message are read-only
        and separate types from those in the ``discord.ui`` namespace.
        In order to modify and edit message components they must be
        converted into a :class:`View` or :class:`LayoutView` first.

        If the message has any v2 components, then you must use
        :class:`LayoutView` in order for them to be converted into
        their respective items. :class:`View` does not support v2 components.

        Parameters
        -----------
        message: :class:`discord.Message`
            The message with components to convert into a view.
        timeout: Optional[:class:`float`]
            The timeout of the converted view.

        Returns
        -------
        Union[:class:`View`, :class:`LayoutView`]
            The converted view. This will always return one of :class:`View` or
            :class:`LayoutView`, and not one of its subclasses.
        """

        if issubclass(cls, View):
            view_cls = View
        elif issubclass(cls, LayoutView):
            view_cls = LayoutView
        else:
            raise TypeError('unreachable exception')

        view = view_cls(timeout=timeout)
        row = 0

        for component in message.components:
            if not view._is_layout() and isinstance(component, ActionRowComponent):
                for child in component.children:
                    item = _component_to_item(child)
                    item.row = row
                    # this error should never be raised, because ActionRows can only
                    # contain items that View accepts, but check anyways
                    if item._is_v2():
                        raise ValueError(f'{item.__class__.__name__} cannot be added to {view.__class__.__name__}')
                    view.add_item(item)
                row += 1
                continue

            item = _component_to_item(component)
            item.row = row

            if item._is_v2() and not view._is_layout():
                raise ValueError(f'{item.__class__.__name__} cannot be added to {view.__class__.__name__}')

            view.add_item(item)
            row += 1

        return view

    def add_item(self, item: Item[Any]) -> Self:
        """Adds an item to the view.

        This function returns the class instance to allow for fluent-style
        chaining.

        Parameters
        -----------
        item: :class:`Item`
            The item to add to the view.

        Raises
        --------
        TypeError
            An :class:`Item` was not passed.
        ValueError
            Maximum number of children has been exceeded, the
            row the item is trying to be added to is full or the item
            you tried to add is not allowed in this View.
        """

        if not isinstance(item, Item):
            raise TypeError(f'expected Item not {item.__class__.__name__}')

        item._update_view(self)
        self._add_count(item._total_count)
        self._children.append(item)
        return self

    def remove_item(self, item: Item[Any]) -> Self:
        """Removes an item from the view.

        This function returns the class instance to allow for fluent-style
        chaining.

        Parameters
        -----------
        item: :class:`Item`
            The item to remove from the view.
        """

        try:
            self._children.remove(item)
        except ValueError:
            pass
        else:
            self._add_count(-item._total_count)
            item._update_view(None)

        return self

    def clear_items(self) -> Self:
        """Removes all items from the view.

        This function returns the class instance to allow for fluent-style
        chaining.
        """
        for child in self._children:
            child._update_view(None)

        self._children.clear()
        self._total_children = 0
        return self

    def find_item(self, id: int, /) -> Optional[Item[Self]]:
        """Gets an item with :attr:`Item.id` set as ``id``, or ``None`` if
        not found.

        .. warning::

            This is **not the same** as ``custom_id``.

        .. versionadded:: 2.6

        Parameters
        ----------
        id: :class:`int`
            The ID of the component.

        Returns
        -------
        Optional[:class:`Item`]
            The item found, or ``None``.
        """
        return _utils_get(self.walk_children(), id=id)

    async def interaction_check(self, interaction: Interaction[ClientT], /) -> bool:
        """|coro|

        A callback that is called when an interaction happens within the view
        that checks whether the view should process item callbacks for the interaction.

        This is useful to override if, for example, you want to ensure that the
        interaction author is a given user.

        The default implementation of this returns ``True``.

        .. note::

            If an exception occurs within the body then the check
            is considered a failure and :meth:`on_error` is called.

        Parameters
        -----------
        interaction: :class:`~discord.Interaction`
            The interaction that occurred.

        Returns
        ---------
        :class:`bool`
            Whether the view children's callbacks should be called.
        """
        return True

    async def on_timeout(self) -> None:
        """|coro|

        A callback that is called when a view's timeout elapses without being explicitly stopped.
        """
        pass

    async def on_error(self, interaction: Interaction[ClientT], error: Exception, item: Item[Any], /) -> None:
        """|coro|

        A callback that is called when an item's callback or :meth:`interaction_check`
        fails with an error.

        The default implementation logs to the library logger.

        Parameters
        -----------
        interaction: :class:`~discord.Interaction`
            The interaction that led to the failure.
        error: :class:`Exception`
            The exception that was raised.
        item: :class:`Item`
            The item that failed the dispatch.
        """
        _log.error('Ignoring exception in view %r for item %r', self, item, exc_info=error)

    async def _scheduled_task(self, item: Item[Any], interaction: Interaction[ClientT]):
        try:
            item._refresh_state(interaction, interaction.data)  # type: ignore

            allow = await item._run_checks(interaction) and await self.interaction_check(interaction)
            if not allow:
                return

            if self.timeout:
                self.__timeout_expiry = time.monotonic() + self.timeout

            await item.callback(interaction)
        except Exception as e:
            return await self.on_error(interaction, e, item)

    def _start_listening_from_store(self, store: ViewStore) -> None:
        self.__cancel_callback = partial(store.remove_view)  # type: ignore
        if self.timeout:
            if self.__timeout_task is not None:
                self.__timeout_task.cancel()

            self.__timeout_expiry = time.monotonic() + self.timeout
            self.__timeout_task = asyncio.create_task(self.__timeout_task_impl())

    def _dispatch_timeout(self):
        if self.__stopped is None or self.__stopped.done():
            return

        if self.__cancel_callback:
            self.__cancel_callback(self)
            self.__cancel_callback = None

        self.__stopped.set_result(True)
        asyncio.create_task(self.on_timeout(), name=f'discord-ui-view-timeout-{self.id}')

    def _dispatch_item(self, item: Item[Any], interaction: Interaction[ClientT]) -> Optional[asyncio.Task[None]]:
        if self.__stopped is None or self.__stopped.done():
            return None

        return asyncio.create_task(self._scheduled_task(item, interaction), name=f'discord-ui-view-dispatch-{self.id}')

    def _refresh(self, components: List[Component]) -> None:
        # fmt: off
        old_state: Dict[str, Item[Any]] = {
            item.custom_id: item  # type: ignore
            for item in self.walk_children()
            if item.is_dispatchable()
        }
        # fmt: on

        for component in _walk_all_components(components):
            custom_id = getattr(component, 'custom_id', None)
            if custom_id is None:
                continue

            try:
                older = old_state[custom_id]
            except KeyError:
                _log.debug('View interaction referenced an unknown item custom_id %s. Discarding', custom_id)
                continue
            else:
                older._refresh_component(component)

    def stop(self) -> None:
        """Stops listening to interaction events from this view.

        This operation cannot be undone.
        """
        if self.__stopped is not None and not self.__stopped.done():
            self.__stopped.set_result(False)

        self.__timeout_expiry = None
        if self.__timeout_task is not None:
            self.__timeout_task.cancel()
            self.__timeout_task = None

        if self.__cancel_callback:
            self.__cancel_callback(self)
            self.__cancel_callback = None

    def is_finished(self) -> bool:
        """:class:`bool`: Whether the view has finished interacting."""
        if self.__stopped is None:
            return False

        return self.__stopped.done()

    def is_dispatching(self) -> bool:
        """:class:`bool`: Whether the view has been added for dispatching purposes."""
        return self.__cancel_callback is not None

    def is_persistent(self) -> bool:
        """:class:`bool`: Whether the view is set up as persistent.

        A persistent view has all their components with a set ``custom_id`` and
        a :attr:`timeout` set to ``None``.
        """
        return self.timeout is None and all(item.is_persistent() for item in self._children)

    async def wait(self) -> bool:
        """|coro|

        Waits until the view has finished interacting.

        A view is considered finished when :meth:`stop` is called
        or it times out.

        Returns
        --------
        :class:`bool`
            If ``True``, then the view timed out. If ``False`` then
            the view finished normally.
        """
        if self.__stopped is None:
            self.__stopped = asyncio.get_running_loop().create_future()

        return await self.__stopped

    def walk_children(self) -> Generator[Item[Any], None, None]:
        """An iterator that recursively walks through all the children of this view
        and its children, if applicable.

        .. versionadded:: 2.6

        Yields
        ------
        :class:`Item`
            An item in the view.
        """

        for child in self.children:
            yield child

            if child._has_children():
                yield from child.walk_children()  # type: ignore


class View(BaseView):
    """Represents a UI view.

    This object must be inherited to create a UI within Discord.

    .. versionadded:: 2.0

    Parameters
    -----------
    timeout: Optional[:class:`float`]
        Timeout in seconds from last interaction with the UI before no longer accepting input.
        If ``None`` then there is no timeout.
    """

    __discord_ui_view__: ClassVar[bool] = True

    if TYPE_CHECKING:

        @classmethod
        def from_message(cls, message: Message, /, *, timeout: Optional[float] = 180.0) -> View: ...

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        children: Dict[str, ItemLike] = {}
        for base in reversed(cls.__mro__):
            for name, member in base.__dict__.items():
                if hasattr(member, '__discord_ui_model_type__'):
                    children[name] = member
                elif isinstance(member, Item) and member._is_v2():
                    raise ValueError(f'{name} cannot be added to this View')

        if len(children) > 25:
            raise TypeError('View cannot have more than 25 children')

        cls.__view_children_items__ = children

    def __init__(self, *, timeout: Optional[float] = 180.0):
        super().__init__(timeout=timeout)
        self.__weights = _ViewWeights(self._children)

    def to_components(self) -> List[Dict[str, Any]]:
        def key(item: Item) -> int:
            return item._rendered_row or 0

        children = sorted(self._children, key=key)
        components: List[Dict[str, Any]] = []
        for _, group in groupby(children, key=key):
            children = [item.to_component_dict() for item in group]
            if not children:
                continue

            components.append(
                {
                    'type': 1,
                    'components': children,
                }
            )

        return components

    def add_item(self, item: Item[Any]) -> Self:
        if len(self._children) >= 25:
            raise ValueError('maximum number of children exceeded')

        if item._is_v2():
            raise ValueError('v2 items cannot be added to this view')

        super().add_item(item)
        try:
            self.__weights.add_item(item)
        except ValueError as e:
            # if the item has no space left then remove it from _children
            self._children.remove(item)
            raise e

        return self

    def remove_item(self, item: Item[Any]) -> Self:
        try:
            self._children.remove(item)
        except ValueError:
            pass
        else:
            self.__weights.remove_item(item)
            item._update_view(None)

        return self

    def clear_items(self) -> Self:
        super().clear_items()
        self.__weights.clear()
        return self


class LayoutView(BaseView):
    """Represents a layout view for components.

    This object must be inherited to create a UI within Discord.

    This differs from a :class:`View` in that it supports all component types
    and uses what Discord refers to as "v2 components".

    You can find usage examples in the :resource:`repository <examples>`

    .. versionadded:: 2.6

    Parameters
    ----------
    timeout: Optional[:class:`float`]
        Timeout in seconds from last interaction with the UI before no longer accepting input.
        If ``None`` then there is no timeout.
    """

    if TYPE_CHECKING:

        @classmethod
        def from_message(cls, message: Message, /, *, timeout: Optional[float] = 180.0) -> LayoutView: ...

    def __init__(self, *, timeout: Optional[float] = 180.0) -> None:
        super().__init__(timeout=timeout)

        if self._total_children > 40:
            raise ValueError('maximum number of children exceeded (40)')

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        children: Dict[str, ItemLike] = {}
        callback_children: Dict[str, ItemCallbackType[Any, Any]] = {}

        for base in reversed(cls.__mro__):
            for name, member in base.__dict__.items():
                if isinstance(member, Item):
                    if member._parent is not None:
                        continue

                    member._rendered_row = member._row
                    children[name] = member
                elif hasattr(member, '__discord_ui_model_type__') and getattr(member, '__discord_ui_parent__', None):
                    callback_children[name] = member

        children.update(callback_children)
        cls.__view_children_items__ = children

    def _is_layout(self) -> bool:
        return True

    def _add_count(self, value: int) -> None:
        if self._total_children + value > 40:
            raise ValueError('maximum number of children exceeded (40)')

        self._total_children = max(0, self._total_children + value)

    def to_components(self):
        components: List[Dict[str, Any]] = []
        for i in self._children:
            components.append(i.to_component_dict())

        return components

    def add_item(self, item: Item[Any]) -> Self:
        if self._total_children >= 40:
            raise ValueError('maximum number of children exceeded (40)')
        super().add_item(item)
        return self

    def content_length(self) -> int:
        """:class:`int`: Returns the total length of all text content in the view's items.

        A view is allowed to have a maximum of 4000 display characters across all its items.
        """
        from .text_display import TextDisplay

        return sum(len(item.content) for item in self.walk_children() if isinstance(item, TextDisplay))


class ViewStore:
    def __init__(self, state: ConnectionState):
        # entity_id: {(component_type, custom_id): Item}
        self._views: Dict[Optional[int], Dict[Tuple[int, str], Item[BaseView]]] = {}
        # message_id: View
        self._synced_message_views: Dict[int, BaseView] = {}
        # custom_id: Modal
        self._modals: Dict[str, Modal] = {}
        # component_type is the key
        self._dynamic_items: Dict[re.Pattern[str], Type[DynamicItem[Item[Any]]]] = {}
        self._state: ConnectionState = state
        self.__tasks: Set[asyncio.Task[None]] = set()

    @property
    def persistent_views(self) -> Sequence[BaseView]:
        # fmt: off
        views = {
            item.view.id: item.view
            for items in self._views.values()
            for item in items.values()
            if item.view and item.view.is_persistent()
        }
        # fmt: on
        return list(views.values())

    def add_task(self, task: asyncio.Task[None]) -> None:
        self.__tasks.add(task)
        task.add_done_callback(self.__tasks.discard)

    def add_dynamic_items(self, *items: Type[DynamicItem[Item[Any]]]) -> None:
        for item in items:
            pattern = item.__discord_ui_compiled_template__
            self._dynamic_items[pattern] = item

    def remove_dynamic_items(self, *items: Type[DynamicItem[Item[Any]]]) -> None:
        for item in items:
            pattern = item.__discord_ui_compiled_template__
            self._dynamic_items.pop(pattern, None)

    def add_view(self, view: BaseView, message_id: Optional[int] = None) -> None:
        view._start_listening_from_store(self)
        if view.__discord_ui_modal__:
            self._modals[view.custom_id] = view  # type: ignore
            return

        dispatch_info = self._views.get(message_id, {})
        is_fully_dynamic = True
        for item in view.walk_children():
            if isinstance(item, DynamicItem):
                pattern = item.__discord_ui_compiled_template__
                self._dynamic_items[pattern] = item.__class__
            elif item.is_dispatchable():
                dispatch_info[(item.type.value, item.custom_id)] = item  # type: ignore
                is_fully_dynamic = False

        view._cache_key = message_id
        if dispatch_info:
            self._views[message_id] = dispatch_info

        if message_id is not None and not is_fully_dynamic:
            self._synced_message_views[message_id] = view

    def remove_view(self, view: BaseView) -> None:
        if view.__discord_ui_modal__:
            self._modals.pop(view.custom_id, None)  # type: ignore
            return

        dispatch_info = self._views.get(view._cache_key)
        if dispatch_info:
            for item in view.walk_children():
                if isinstance(item, DynamicItem):
                    pattern = item.__discord_ui_compiled_template__
                    self._dynamic_items.pop(pattern, None)
                elif item.is_dispatchable():
                    dispatch_info.pop((item.type.value, item.custom_id), None)  # type: ignore

        if dispatch_info is not None and len(dispatch_info) == 0:
            self._views.pop(view._cache_key, None)

        self._synced_message_views.pop(view._cache_key, None)  # type: ignore

    async def schedule_dynamic_item_call(
        self,
        component_type: int,
        factory: Type[DynamicItem[Item[Any]]],
        interaction: Interaction[ClientT],
        custom_id: str,
        match: re.Match[str],
    ) -> None:
        if interaction.message is None:
            return

        view_cls = View if not interaction.message.flags.components_v2 else LayoutView
        view = view_cls.from_message(interaction.message, timeout=None)

        base_item = _utils_find(
            lambda i: i.type.value == component_type and getattr(i, 'custom_id', None) == custom_id,
            view.walk_children(),
        )

        # if the item is not found then return
        if not base_item:
            return

        try:
            item = await factory.from_custom_id(interaction, base_item, match)
        except Exception:
            _log.exception('Ignoring exception in dynamic item creation for %r', factory)
            return

        # Swap the item in the view or parent with our new dynamic item
        # Prioritize the item parent:
        parent = base_item._parent or view

        try:
            parent._swap_item(base_item, item, custom_id)
        except ValueError:
            return

        item._view = view
        item._rendered_row = base_item._rendered_row
        item._refresh_state(interaction, interaction.data)  # type: ignore

        try:
            allow = await item.interaction_check(interaction)
        except Exception:
            allow = False

        if not allow:
            return

        try:
            await item.callback(interaction)
        except Exception:
            _log.exception('Ignoring exception in dynamic item callback for %r', item)

    def dispatch_dynamic_items(self, component_type: int, custom_id: str, interaction: Interaction[ClientT]) -> None:
        for pattern, item in self._dynamic_items.items():
            match = pattern.fullmatch(custom_id)
            if match is not None:
                self.add_task(
                    asyncio.create_task(
                        self.schedule_dynamic_item_call(component_type, item, interaction, custom_id, match),
                        name=f'discord-ui-dynamic-item-{item.__name__}-{custom_id}',
                    )
                )

    def dispatch_view(self, component_type: int, custom_id: str, interaction: Interaction[ClientT]) -> None:
        self.dispatch_dynamic_items(component_type, custom_id, interaction)
        interaction_id: Optional[int] = None
        message_id: Optional[int] = None
        # Realistically, in a component based interaction the Interaction.message will never be None
        # However, this guard is just in case Discord screws up somehow
        msg = interaction.message
        if msg is not None:
            message_id = msg.id
            if msg.interaction_metadata:
                interaction_id = msg.interaction_metadata.id

        key = (component_type, custom_id)

        # The entity_id can either be message_id, interaction_id, or None in that priority order.
        item: Optional[Item[BaseView]] = None
        if message_id is not None:
            item = self._views.get(message_id, {}).get(key)

        if item is None and interaction_id is not None:
            try:
                items = self._views.pop(interaction_id)
            except KeyError:
                item = None
            else:
                item = items.get(key)
                # If we actually got the items, then these keys should probably be moved
                # to the proper message_id instead of the interaction_id as they are now.
                # An interaction_id is only used as a temporary stop gap for
                # InteractionResponse.send_message so multiple view instances do not
                # override each other.
                # NOTE: Fix this mess if /callback endpoint ever gets proper return types
                self._views.setdefault(message_id, {}).update(items)

        if item is None:
            # Fallback to None message_id searches in case a persistent view
            # was added without an associated message_id
            item = self._views.get(None, {}).get(key)

        # If 3 lookups failed at this point then just discard it
        if item is None:
            return

        if item.view is None:
            _log.warning('View interaction referencing unknown view for item %s. Discarding', item)
            return

        task = item.view._dispatch_item(item, interaction)
        if task is not None:
            self.add_task(task)

    def dispatch_modal(
        self,
        custom_id: str,
        interaction: Interaction[ClientT],
        components: List[ModalSubmitComponentInteractionDataPayload],
        resolved: ResolvedDataPayload,
    ) -> None:
        modal = self._modals.get(custom_id)
        if modal is None:
            _log.debug('Modal interaction referencing unknown custom_id %s. Discarding', custom_id)
            return

        self.add_task(modal._dispatch_submit(interaction, components, resolved))

    def remove_interaction_mapping(self, interaction_id: int) -> None:
        # This is called before re-adding the view
        self._views.pop(interaction_id, None)
        self._synced_message_views.pop(interaction_id, None)

    def is_message_tracked(self, message_id: int) -> bool:
        return message_id in self._synced_message_views

    def remove_message_tracking(self, message_id: int) -> Optional[BaseView]:
        return self._synced_message_views.pop(message_id, None)

    def update_from_message(self, message_id: int, data: List[ComponentBasePayload]) -> None:
        components: List[Component] = []

        for component_data in data:
            component = _component_factory(component_data, self._state)  # type: ignore

            if component is not None:
                components.append(component)

        # pre-req: is_message_tracked == true
        view = self._synced_message_views[message_id]
        view._refresh(components)
